(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis myclassical.
Require Import lang_common lang_spec lang_mid lang_tgt.
Require Import model_common model_untyped model_properties_untyped.
Require Import gw_common gw_untyped codegen codegen_correct_aux.
Require Import ssr util.

Unset Implicit Arguments.       
Global Set Bullet Behavior "Strict Subproofs".



Lemma tr_exp_correct_unit: forall G,
  tm_bsim G (fun n => map enc (tr_exp IEunit n (fvars G) (tvars G) (kvars G)))
            IEunit.
Proof.
  ii; subst; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]].
  revert CR; tm_prepare'.

  eapply pesim_stepb_obsolete.
  { i; rrdes XTR RL; clarify.
    destruct ee; s in Ec; clarify.
    esplits.
    - econs 1.
      inst_pairs; eapply IRok.
      eapply IRlet; s; eauto.
    - inst_pairs; rr; esplits; s; noemp; eauto. }

  set (s' := (mrf, P0, s, s3, (fst s4, {mrf r_arg |-> v} ## snd s4))).

  assert (VLD': pmvalid ({mrf r_arg |-> v} ## snd s4)).
  { eapply pmvalid_plus_r, pmvalid_plus_r.
    rewrite 2!(pmplus_swap {_|->_}) in HV; eauto.
  }

  assert (EXT': extpub (wlift mylw) s' (mrf, P0, s, s3, s4)).
  { clear -HV VLD'.
    s; splits; auto.
    i; eapply pmread_plus; eauto using pmplus_comm.
  }

  eapply pesim_anti with (s' := s').
  - rewrite tgt_csplit mid_csplit.
    eapply pesim_release; last first; try by rr; s; eauto.
    { rr; esplits.
      - rr; esplits; eauto.
        + s; rr; esplits; eauto.
        + eapply tm_refconfrel_mon; eauto.
          s; i; eapply gk_mon; eauto.
          s; rr; esplits; eauto.
          i; eapply pmread_plus; eauto.
          rewrite ->pmplus_comm; eauto.
      - rr; esplits; eauto.
      - s; noemp.
        rewrite -pmplus_assoc (pmplus_swap refha) (pmplus_swap hp); auto.
      - noemp; auto. }
    eapply KSIM; auto.
    + s; eapply pmread_plus; eauto.
    + econs; eapply vclgen_unit; s; eauto.
  - apply extpub_in_ext; auto.
Qed.



Lemma tr_exp_correct_nat: forall G n,
  tm_bsim G (fun m => map enc (tr_exp (IEnat n) m (fvars G) (tvars G) (kvars G)))
            (IEnat n).
Proof.
  ii; subst; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]].
  revert CR; tm_prepare'.

  eapply pesim_dance; first (by eauto); clear RLa0.
  - clear SV HV FIN.

    dance_init.
    { eapply Sok; [by vauto|eapply Sld]; [by read_code Eladdr|by eauto]. }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
      replace (mrf0 r_arg) with (add_offset (mrf r_arg) o_zero)
      ; last (by s; unfold mrf; rewrite tmext_neq; vauto); eauto. }
    { depmcplus0; eauto using pmvalid_plus_cell, pmfinite_plus_cell. }

    eapply dance_O; instantiate (1 := fun _ _ => _); s; eauto.

  - i; rrdes XTRb RLb0; clarify.
    destruct ee; s in Ec; clarify.
    esplits; first 2 last; last first.
    + left; econs.
      inst_pairs; eapply IRok.
      eapply IRlet; s; eauto.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.
      set (s' := ([n//r_aux]>>mrf, P0, s, s3, (fst s4, {mrf r_arg |-> n} ## snd s4))).
      assert (VLD': pmvalid ({mrf r_arg |-> n} ## snd s4)).
      { eapply pmvalid_plus_cell, pmvalid_plus_r, pmvalid_plus_r.
        rewrite 2!(pmplus_swap {_|->_}) in HV; eauto.
      }
      assert (EXT': extpub (wlift mylw) s' (mrf, P0, s, s3, s4)).
      { clear -HV VLD'.
        s; splits; tmext_simpl; auto.
        i; eapply pmread_plus; eauto using pmplus_comm.
      }
      exists s'; split; [by apply extpub_in_ext; exact EXT'|].
      rewrite tgt_csplit; apply pesim_release; last first; try by rr; s; eauto.
      { rr; esplits; last by eauto.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto.
            instantiate (1 := {| cpc := Nothing; crf := Just [n//r_aux]>>mrf; cst := freest; chp := hp |}).
            s; rrdes CPa.
            rr; splits; s; tmext_simpl; auto.
          + s; eapply tm_refconfrel_mon
            ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
            ; s; i; eapply gk_mon; eauto
            ; rrdes EXT'; rr; eauto using extpub_in_ext.
        - rr; esplits; eauto.
        - s; noemp.
          rewrite -pmplus_assoc (pmplus_swap refha) (pmplus_swap hp); auto.
      }
      inst_pairs; eapply KSIM; auto.
      * s; tmext_simpl; auto.
      * s; tmext_simpl; eapply pmread_plus; eauto.
      * econs; eapply vclgen_nat; s; eauto.
    + s; noemp; rewrite plus_comm; rr; esplits; eauto.
Qed.



Lemma tr_exp_correct_fst: forall G lx
    (IN: In (inl lx) G),
  tm_bsim G (fun n => map enc (tr_exp (IEfst lx) n (fvars G) (tvars G) (kvars G)))
            (IEfst lx).
Proof.
  ii; subst; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  hexploit (values_for_var RSV RDV IN); i; des.

  revert CR; tm_prepare'.

  destruct (LEM (exists vb' vb'', vb = IVpair vb' vb'')); last first.
  { 
    pfold; econs; i.
    destruct cfb; rrdes XTRb; clarify.
    invc CNF; des; subst; noemp.
    rrdes MB; s in Ec; clarify.
    econs 1.
    - instantiate (1:=(None, (envb, IElet (IEfst lx) (inl x) kb))).
      econs 1; eapply IRerr; vauto; i; invc FH.
      by s in EVAL; rewrite INb in EVAL; destruct vb; clarify; eauto 10.
    - eauto using esim_err.
  }
  des; subst.

  eapply pesim_stepb_obsolete.
  { i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits.
    - econs 1; inst_pairs; eapply IRok.
      eapply IRlet; s; rewrite INb; s; eauto.
    - inst_pairs; s; noemp; rr; esplits; eauto.
  }

  eapply pesim_getVal; eauto.
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr; pmplus_eq'.
  }
  { eapply reprlist_mon; eauto.
    i; eapply pmread_plus; eauto; pmplus_eq'.
  }
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr map_app pmcells_app SORD; pmplus_eq'.
  }
  { eapply wfo_le_n_add. }

  
  dup VCL; apply tm_vclosDb in VCL; des; rrdes VCL.
  dup CPa; rrdes CPa; clean; apply HP in VCL; s in VCL; clear HP ST; des.

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0.
  - clear SV HV FIN; rename mrf into mrf0.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq'.
      - econs; unfold mrf; tmext_simpl; s.
        eapply pmread_plus; eauto; pmplus_eq'.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq'.
      - replace (mrf0 r_arg) with (add_offset (mrf r_arg) o_zero)
      ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { depmcplus0; eauto using pmvalid_plus_cell, pmfinite_plus_cell. }

    eapply dance_O; instantiate (1 := fun _ _ => _); s; eauto.

  - i; rrdes XTRb RLb0; clarify.
    s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.

      setcrf mrf'.
      set (s' := (mrf', P0, s, s3, (fst s4, {mrf r_arg |-> va'} ## snd s4))).
      assert (VLD': pmvalid ({mrf r_arg |-> va'} ## snd s4)).
      { decomp_pmvalid HV; auto. }
      assert (EXT': extpub (wlift mylw) s' (mrf, P0, s, s3, s4)).
      { clear -HV VLD'; unfold s', mrf'.
        s; splits; tmext_simpl; auto.
        i; eapply pmread_plus; eauto; rewrite ->pmplus_comm; eauto.
      }
      exists s'; split; [by apply extpub_in_ext; exact EXT'|].

      rewrite tgt_csplit; apply pesim_release; last first; try by rr; s; eauto.
      { unfold s', mrf'.
        rr; esplits; last by eauto.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto.
            instantiate (1 := mkconf _ _ _ _).
            rrdes CPa0; rr; esplits; s; tmext_simpl; eauto.
          + s; eapply tm_refconfrel_mon
            ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
            ; s; i; eapply gk_mon; eauto
            ; rrdes EXT'; rr; eauto using extpub_in_ext.
        - rr; esplits; eauto.
        - s; noemp; lhs rewrite -pmplus_swap -!pmplus_assoc; eauto.
      }
      
      simpl tr_exp in KSIM; rewrite map_length app_length in KSIM; simpl length in KSIM.
      rewrite -(plus_comm 2) SORD in KSIM; simpl plus in KSIM; rewrite plus_comm in KSIM.
      inst_pairs; eapply KSIM.
      * exact EXT'.
      * unfold s', mrf'; s; tmext_simpl; auto.
      * unfold s', mrf'; s; tmext_simpl; eapply pmread_plus; eauto.
      * eapply vclos_mon; eauto using (gk_mon GK).
    + s; noemp; try rewrite plus_comm; rr; esplits; eauto.
Qed.



Lemma tr_exp_correct_snd: forall G lx
    (IN: In (inl lx) G),
  tm_bsim G (fun n => map enc (tr_exp (IEsnd lx) n (fvars G) (tvars G) (kvars G)))
            (IEsnd lx).
Proof.
  

  ii; subst; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  hexploit (values_for_var RSV RDV IN); i; des.

  revert CR; tm_prepare'.

  destruct (LEM (exists vb' vb'', vb = IVpair vb' vb'')); last first.
  { 
    pfold; econs; i.
    destruct cfb; rrdes XTRb; clarify.
    invc CNF; des; subst; noemp.
    rrdes MB; s in Ec; clarify.
    econs 1.
    - instantiate (1:=(None, (envb, IElet (IEsnd lx) (inl x) kb))).
      econs 1; eapply IRerr; vauto; i; invc FH.
      s in EVAL; rewrite INb in EVAL; destruct vb; clarify; eauto 10.
    - eauto using esim_err.
  }
  des; subst.

  eapply pesim_stepb_obsolete.
  { i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits.
    - econs 1; inst_pairs; eapply IRok.
      eapply IRlet; s; rewrite INb; s; eauto.
    - inst_pairs; s; noemp; rr; esplits; eauto.
  }

  eapply pesim_getVal; eauto.
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr; pmplus_eq'.
  }
  { eapply reprlist_mon; eauto.
    i; eapply pmread_plus; eauto; pmplus_eq'.
  }
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr map_app pmcells_app SORD; pmplus_eq'.
  }
  { eapply wfo_le_n_add. }

  
  dup VCL; apply tm_vclosDb in VCL; des; rrdes VCL.
  dup CPa; rrdes CPa; clean; apply HP in VCL; s in VCL; clear HP ST; des.

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0.
  - clear SV HV FIN; rename mrf into mrf0.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - econs; unfold mrf; tmext_simpl; s.
        eapply pmread_plus; eauto; pmplus_eq'.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - replace (mrf0 r_arg) with (add_offset (mrf r_arg) o_zero)
      ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { depmcplus0; eauto using pmvalid_plus_cell, pmfinite_plus_cell. }

    eapply dance_O; instantiate (1 := fun _ _ => _); s; eauto.

  - i; rrdes XTRb RLb0; clarify.
    s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.

      setcrf mrf'.
      set (s' := (mrf', P0, s, s3, (fst s4, {mrf r_arg |-> va''} ## snd s4))).
      assert (VLD': pmvalid ({mrf r_arg |-> va''} ## snd s4)).
      { decomp_pmvalid HV; auto. }
      assert (EXT': extpub (wlift mylw) s' (mrf, P0, s, s3, s4)).
      { clear -HV VLD'; unfold s', mrf'.
        s; splits; tmext_simpl; auto.
        i; eapply pmread_plus; eauto; rewrite ->pmplus_comm; eauto.
      }
      exists s'; split; [by apply extpub_in_ext; exact EXT'|].

      rewrite tgt_csplit; apply pesim_release; last first; try by rr; s; eauto.
      { unfold s', mrf'.
        rr; esplits; last by eauto.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto.
            instantiate (1 := mkconf _ _ _ _).
            rrdes CPa0; rr; esplits; s; tmext_simpl; eauto.
          + s; eapply tm_refconfrel_mon
            ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
            ; s; i; eapply gk_mon; eauto
            ; rrdes EXT'; rr; eauto using extpub_in_ext.
        - rr; esplits; eauto.
        - s; noemp; lhs rewrite -pmplus_swap -!pmplus_assoc; eauto.
      }
      
      simpl tr_exp in KSIM; rewrite map_length app_length in KSIM; simpl length in KSIM.
      rewrite -(plus_comm 2) SORD in KSIM; simpl plus in KSIM; rewrite plus_comm in KSIM.
      inst_pairs; eapply KSIM.
      * exact EXT'.
      * unfold s', mrf'; s; tmext_simpl; auto.
      * unfold s', mrf'; s; tmext_simpl; eapply pmread_plus; eauto.
      * eapply vclos_mon; eauto using (gk_mon GK).
    + s; noemp; try rewrite plus_comm; rr; esplits; eauto.
Qed.



Lemma tr_exp_correct_inl: forall G lx
    (IN: In (inl lx) G),
  tm_bsim G (fun n => map enc (tr_exp (IEinl lx) n (fvars G) (tvars G) (kvars G)))
            (IEinl lx).
Proof.
  ii; subst; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  
  hexploit (values_for_var RSV RDV IN); i; des.

  revert CR; tm_prepare'.
  eapply pesim_stepb_obsolete.
  


  { i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits.
    - econs 1; inst_pairs; eapply IRok.
      eapply IRlet; s; rewrite INb; s; eauto.
    - inst_pairs; s; noemp; rr; esplits; eauto.
  }

  eapply pesim_getVal; eauto.
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr; pmplus_eq'.
  }
  { eapply reprlist_mon; eauto.
    i; eapply pmread_plus; eauto; pmplus_eq'.
  }
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr map_app pmcells_app SORD; pmplus_eq'.
  }
  { eapply wfo_le_n_add. }

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0.
  - clear SV HV FIN.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Salc].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto. }
    { depmcplus0. }
    tmext_uniq.

    rename mrf into mrf0.       
    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - replace newloc with (add_offset (mrf r_clo) o_zero)
        ; last (by s; unfold mrf; tmext_simpl; auto).
        rewrite pmcells_cons -pmplus_assoc; eauto.
    }
    { pmcells_fold; fold newloc; depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD.
        s; rewrite ->?pmcells_cons; pmplus_eq''.
      - rewrite !pmcells_cons; replace (S newloc) with (add_offset (mrf r_clo) o_one)
        ; last (by s; unfold mrf; tmext_simpl; auto).
        rewrite -!pmplus_assoc (pmplus_swap {_|->_}); eauto.
    }
    { s; tmext_simpl; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD.
        s; rewrite ->?pmcells_cons; pmplus_eq''.
      - replace (mrf0 r_arg) with (add_offset (mrf r_arg) o_zero)
        ; last (by s; unfold mrf; tmext_simpl; auto).
        rewrite (pmplus_swap {_|=>_}); eauto.
    }
    { rewrite (pmplus_swap {_|=>_}) in HV0, FIN0.
      depmcplus0; eauto using pmvalid_plus_cell, pmfinite_plus_cell.
    }

    eapply dance_O; instantiate (1 := fun _ cf _ _ _ _ => depmach1 cf); s; eauto.

  - i; rrdes XTRb RLb0; clarify.
    destruct ee; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_dep 1 RLa XTRa cfas; frames_nodep RLa cfas.
      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.
      setnewloc.
      set (s' := ([0//r_aux]>>[newloc//r_clo]>>[va//r_ret]>>mrf,
                  extend P0 (ts_inl va) newloc, s, s3,
                  (fst s4, {mrf r_arg |-> newloc} ## snd s4))).
      
      assert (VLD': pmvalid ({mrf r_arg |-> newloc} ## snd s4)).
      { clear -HV.
        rewrite <-(pmplus_comm (snd s4)), <-2(pmplus_swap (snd s4)) in HV.
        eapply pmvalid_plus_l; rewrite <-pmplus_assoc; eauto.
      }
      assert (EXT': extpub (wlift mylw) s' (mrf, P0, s, s3, s4)).
      { clear -HV VLD'.
        s; splits; tmext_simpl; auto; [by i; right; eauto|].
        i; eapply pmread_plus; eauto.
        rewrite ->pmplus_comm; eauto. 
      }
      exists s'; split; [by apply extpub_in_ext; exact EXT'|].
      
      rewrite tgt_csplit; apply pesim_release; last first; try by rr; s; eauto.
      { unfold s'.
        rr; esplits; last by eauto.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto.
            instantiate (1 := {| cpc := Nothing; crf := Just [0//r_aux]>>[newloc//r_clo]>>[va//r_ret]>>mrf;
                                 cst := freest; chp := {newloc |=> [0; va]} ## hp |}).
            s; rrdes CPa; s; s in HP.
            rr; splits; s; tmext_simpl; auto.
            unfold extend; i; des.
            * rewrite -IN1 -IN0; s; split. 
              { eapply pmread_plus; eauto.
                - rewrite pmcells_cons; pmplus_eq'.
                - decomp_pmvalid HV; eauto. }
              { eapply pmread_plus; eauto.
                - rewrite !pmcells_cons; pmplus_eq'.
                - decomp_pmvalid HV; eauto. }
            * eapply repr_mon; [s; i|by eauto].
              eapply pmread_plus; try rewrite ->pmplus_comm; eauto.
              decomp_pmvalid HV; rewrite pmplus_comm; eauto.
              
          + s; eapply tm_refconfrel_mon; s; unfold extend
            ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
            ; s; i; eapply gk_mon; eauto
            ; rrdes EXT'; rr; eauto using extpub_in_ext.
        - rr; esplits; eauto.
        - s; noemp.
          rewrite pmplus_swap -!pmplus_assoc; eauto. }
      inst_pairs; hexploit KSIM.
      * exact EXT'.
      * s; tmext_simpl; auto.
      * s; tmext_simpl; eapply pmread_plus; eauto.
      * econs; eapply vclgen_inl; s; unfold extend; eauto.
        eapply vclos_mon; eauto using (gk_mon GK).
      * s; rewrite map_length app_length; s; rewrite -(plus_comm 6) SORD.
        s; rewrite plus_comm; eauto.
        
    + s; noemp; try rewrite plus_comm; rr; esplits; eauto.
Qed.



Lemma tr_exp_correct_inr: forall G lx
    (IN: In (inl lx) G),
  tm_bsim G (fun n => map enc (tr_exp (IEinr lx) n (fvars G) (tvars G) (kvars G)))
            (IEinr lx).
Proof.
  

  ii; subst; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  
  hexploit (values_for_var RSV RDV IN); i; des.

  revert CR; tm_prepare'.
  eapply pesim_stepb_obsolete.
  


  { i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits.
    - econs 1; inst_pairs; eapply IRok.
      eapply IRlet; s; rewrite INb; s; eauto.
    - inst_pairs; s; noemp; rr; esplits; eauto.
  }

  eapply pesim_getVal; eauto.
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr; pmplus_eq'.
  }
  { eapply reprlist_mon; eauto.
    i; eapply pmread_plus; eauto; pmplus_eq'.
  }
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr map_app pmcells_app SORD; pmplus_eq'.
  }
  { eapply wfo_le_n_add.
  }

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0.
  - clear SV HV FIN.            
    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Salc].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto. }
    { depmcplus0. }
    tmext_uniq.

    rename mrf into mrf0.
    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - replace newloc with (add_offset (mrf r_clo) o_zero)
        ; last (by s; unfold mrf; tmext_simpl; auto).
        rewrite pmcells_cons -pmplus_assoc; eauto. }
    { pmcells_fold; fold newloc.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD.
        s; rewrite ->?pmcells_cons; pmplus_eq''.
      - rewrite !pmcells_cons; replace (S newloc) with (add_offset (mrf r_clo) o_one)
        ; last (by s; unfold mrf; tmext_simpl; auto).
        rewrite -!pmplus_assoc (pmplus_swap {_|->_}); eauto.
    }
    { s; tmext_simpl; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD.
        s; rewrite ->?pmcells_cons; pmplus_eq''.
      - replace (mrf0 r_arg) with (add_offset (mrf r_arg) o_zero)
        ; last (by s; unfold mrf; tmext_simpl; auto).
        rewrite (pmplus_swap {_|=>_}); eauto.
    }
    { rewrite (pmplus_swap {_|=>_}) in HV0, FIN0.
      depmcplus0; eauto using pmvalid_plus_cell, pmfinite_plus_cell.
    }

    eapply dance_O; instantiate (1 := fun _ cf _ _ _ _ => depmach1 cf); s; eauto.

  - i; rrdes XTRb RLb0; clarify.
    destruct ee; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_dep 1 RLa XTRa cfas; frames_nodep RLa cfas.
      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.
      setnewloc.
      set (s' := ([1//r_aux]>>[newloc//r_clo]>>[va//r_ret]>>mrf,
                  extend P0 (ts_inr va) newloc, s, s3,
                  (fst s4, {mrf r_arg |-> newloc} ## snd s4))).
      
      assert (VLD': pmvalid ({mrf r_arg |-> newloc} ## snd s4)).
      { clear -HV.
        rewrite <-(pmplus_comm (snd s4)), <-2(pmplus_swap (snd s4)) in HV.
        eapply pmvalid_plus_l; rewrite <-pmplus_assoc; eauto.
      }
      assert (EXT': extpub (wlift mylw) s' (mrf, P0, s, s3, s4)).
      { clear -HV VLD'.
        s; splits; tmext_simpl; auto; [by i; right; eauto|].
        i; eapply pmread_plus; eauto.
        rewrite ->pmplus_comm; eauto.
      }
      exists s'; split; [by apply extpub_in_ext; exact EXT'|].
      
      rewrite tgt_csplit; apply pesim_release; last first; try by rr; s; eauto.
      { unfold s'.
        rr; esplits; last by eauto.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto.
            instantiate (1 := {| cpc := Nothing; crf := Just [1//r_aux]>>[newloc//r_clo]>>[va//r_ret]>>mrf;
                                 cst := freest; chp := {newloc |=> [1; va]} ## hp |}).
            s; rrdes CPa; s; s in HP.
            rr; splits; s; tmext_simpl; auto.
            unfold extend; i; des. 
            * rewrite -IN1 -IN0; s; esplits; first 1 last.
              { eapply pmread_plus; eauto.
                - rewrite pmcells_cons; pmplus_eq'.
                - decomp_pmvalid HV; eauto. }
              { eapply pmread_plus; eauto.
                - rewrite !pmcells_cons; pmplus_eq'.
                - decomp_pmvalid HV; eauto. }
              { eauto. }
            * eapply repr_mon; [s; i|by eauto].
              eapply pmread_plus; try rewrite ->pmplus_comm; eauto.
              decomp_pmvalid HV; rewrite pmplus_comm; eauto.
              
          + s; eapply tm_refconfrel_mon; s; unfold extend
            ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
            ; s; i; eapply gk_mon; eauto
            ; rrdes EXT'; rr; eauto using extpub_in_ext.
        - rr; esplits; eauto.
        - s; noemp.
          rewrite pmplus_swap -!pmplus_assoc; eauto. }
      inst_pairs; hexploit KSIM.
      * exact EXT'.
      * s; tmext_simpl; auto.
      * s; tmext_simpl; eapply pmread_plus; eauto.
      * econs; eapply vclgen_inr; s; unfold extend; eauto.
        eapply vclos_mon; eauto using (gk_mon GK).
      * s; rewrite map_length app_length; s; rewrite -(plus_comm 6).
        s; rewrite plus_comm SORD; eauto.
        
    + s; noemp; try rewrite plus_comm; rr; esplits; eauto.
Qed.



Lemma tr_exp_correct_pair: forall G lx ly
    (IN: In (inl lx) G)
    (IN: In (inl ly) G),
  tm_bsim G (fun n => map enc (tr_exp (IEpair lx ly) n (fvars G) (tvars G) (kvars G)))
            (IEpair lx ly).
Proof.
  ii; subst; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  hexploit (values_for_var RSV RDV IN); i; des.
  hexploit (values_for_var RSV RDV IN0); i; des.

  revert CR; tm_prepare'.
  eapply pesim_stepb_obsolete.
  { i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits.
    - econs 1; inst_pairs; eapply IRok.
      eapply IRlet; s; rewrite INb INb0; s; eauto.
    - inst_pairs; s; noemp; rr; esplits; eauto.
  }

  move INa at bottom.
  eapply pesim_getVal; eauto.
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr; pmplus_eq'.
  }
  { eapply reprlist_mon; eauto.
    i; eapply pmread_plus; eauto.
    pmplus_eq'.
  }
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr map_app pmcells_app SORD; pmplus_eq'.
  }
  { eapply wfo_le_n_add.
  }

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0.
  - clear SV HV FIN.
    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Salc].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    rename mrf into mrf0.
    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length SORD; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - replace newloc with (add_offset (mrf r_clo) o_zero)
        ; last (by s; unfold mrf; tmext_simpl; auto).
        rewrite pmcells_cons -pmplus_assoc; eauto.
    }
    { s; tmext_simpl; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    eapply dance_O; instantiate (1 := fun _ cf _ => depmach1 cf); s; eauto.

  - i; rrdes XTRb RLb0; clarify.
    destruct ee; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_dep 1 RLa XTRa cfas; frames_nodep RLa cfas.
      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setnewloc.
      esplits; eauto.
      clear INa; eapply pesim_getVal; eauto.
      { instantiate (1 := 3 + n + length (getVal lx n (map fst senva) (tvars G) r_ret)).
        i; eapply pmread_plus; eauto.
        rewrite Eladdr SORD.
        replace (_ - _) with (mpc - n); last first.
        - clear; set (x := length _); omega.
        - pmplus_eq'.
      }
      { tmext_simpl; eapply reprlist_mon; eauto.
        i; eapply pmread_plus; eauto.
        pmplus_eq'.
      }
      { i; eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app; s.
        rewrite (plus_comm n) !pmcells_cons map_app pmcells_app map_length SORD; pmplus_eq''.
      }
      { eapply wfo_le_n_add. }
      tmext_uniq.

      eapply pesim_dance
      ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
      ; clear RLa0.
      * clear SV HV FIN FIN0; rename mrf into mrf0.
        dance_init.
        { eapply Sok; [s; i; omega|eapply Ssto].
          - eapply pmread_plus; eauto.
            rewrite Eladdr map_app pmcells_app map_length; s.
            rewrite ->?pmcells_cons, ->map_app, pmcells_app; s.
            rewrite ->?pmcells_cons, map_length, (plus_comm n), SORD; pmplus_eq''.
          - rewrite !pmcells_cons pmcells_nil; noemp.
            replace (S newloc) with (add_offset (mrf r_clo) o_one)
            ; last (by unfold mrf; s; tmext_simpl; auto).
            rewrite <-?pmplus_assoc, <-?(pmplus_swap {add_offset _ _|->_}); eauto.
        }
        { s; tmext_simpl; pmcells_order; pmcells_fold.
          depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
        }

        dance_init.
        { eapply Sok; [s; i; omega|eapply Ssto].
          - eapply pmread_plus; eauto.
            rewrite Eladdr map_app pmcells_app map_length; s.
            rewrite ->?pmcells_cons, ->map_app, pmcells_app; s.
            rewrite ->?pmcells_cons, map_length, (plus_comm n), SORD; pmplus_eq''.
          - replace (mrf0 r_arg) with (add_offset (mrf r_arg) o_zero)
            ; last (by unfold mrf; s; tmext_simpl; auto).
            rewrite <-?pmplus_assoc, <-?(pmplus_swap {add_offset _ _|->_}); eauto.
        }
        { rewrite pmplus_swap in HV0, FIN0.
          depmcplus0; eauto using pmvalid_plus_cell, pmfinite_plus_cell.
        }

        eapply dance_O; instantiate (1 := fun _ _ => _); s; eauto.

      * inst_pairs.
        i; rrdes XTRb RLb0; clarify.
        destruct ee; s in Ec; clarify.
        esplits; first 2 last; last first.
        { right; splits; eauto using wfo_le_n_add. }
        { i; frames_dep 1 RLa XTRa cfas; frames_nodep RLa cfas.
          destruct ea'; clear SV HV FIN.
          destruct_tgt_real RLa.
          
          set (s' := ([newloc//r_clo]>>[2//r_aux]>>[va0//r_ret]>>mrf,
                  extend P0 (ts_pair va va0) newloc, s, s3,
                  (fst s4, {mrf r_arg |-> newloc} ## snd s4))).

          assert (VLD': pmvalid ({mrf r_arg |-> newloc} ## snd s4)).
          { clear -HV.
            rewrite <-(pmplus_comm (snd s4)), <-2(pmplus_swap (snd s4)) in HV.
            eapply pmvalid_plus_l; rewrite <-pmplus_assoc; eauto.
          }
          assert (EXT': extpub (wlift mylw) s' (mrf, P0, s, s3, s4)).
          { clear -HV VLD'.
            s; splits; tmext_simpl; auto; [by i; right; eauto|].
            i; eapply pmread_plus; eauto.
            rewrite ->pmplus_comm; eauto.
          }
          exists s'; split; [by apply extpub_in_ext; exact EXT'|].
          rewrite tgt_csplit; apply pesim_release; last first; try by rr; s; eauto.
          { unfold s'.
            rr; esplits; last by eauto.
            - rr; esplits; eauto.
              + s; rr; esplits; eauto.
                instantiate (1 := {| cpc := Nothing; crf := Just [newloc//r_clo]>>[2//r_aux]>>[va0//r_ret]>>mrf;
                                     cst := freest; chp := {newloc |=> [va; va0]} ## hp |}).
                s; rrdes CPa; s; s in HP.
                rr; splits; s; tmext_simpl; auto.
                unfold extend; i; des.
                * rewrite -IN1 IN2; s; split.
                  { eapply pmread_plus; eauto.
                    - (rhs rewrite pmcells_cons -!pmplus_assoc); eauto.
                    - rewrite -IN2; decomp_pmvalid HV; eauto.
                  }
                  { eapply pmread_plus; eauto.
                    - (rhs rewrite pmcells_cons -!pmplus_assoc pmplus_swap); noemp; eauto.
                    - rewrite -IN2; decomp_pmvalid HV; eauto.
                  }
                * eapply repr_mon; [s; i|by eauto].
                  eapply pmread_plus; try rewrite ->pmplus_comm; eauto.
                  decomp_pmvalid HV; rewrite pmplus_comm; eauto.
              + s; eapply tm_refconfrel_mon; s; unfold extend
                ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
                ; s; i; eapply gk_mon; eauto
                ; rrdes EXT'; rr; eauto using extpub_in_ext.
            - rr; esplits; eauto.
            - s; noemp.
              rewrite ->2!(@tmext_swap _ _ _ r_ret); try by vauto.
              rewrite pmplus_swap -!pmplus_assoc; eauto.
          }
          inst_pairs; hexploit KSIM.
          - exact EXT'.
          - s; tmext_simpl; auto.
          - s; tmext_simpl; eapply pmread_plus; eauto.
          - econs; eapply vclgen_pair; s; unfold extend; eauto
            ; eapply vclos_mon; eauto using (gk_mon GK).
          - replace (_ + _) with (S (S (S (S (S
                             (mpc + length (getVal lx n (map fst senva) (tvars G) r_ret) +
                              length (getVal ly (S (S (S (n + length (getVal lx n (map fst senva) (tvars G) r_ret))))) (map fst senva) (tvars G) r_ret))))))); eauto.
            s; rewrite SORD map_length app_length; s.
            rewrite app_length (plus_comm n); s; omega.
        }
        { s; noemp.
          instantiate (1 := Just _) in H0. clarify. rewrite -H0. 
          rr; esplits; eauto.
        }
    + s; noemp; try rewrite plus_comm; rr; esplits; eauto.
Qed.



Lemma tr_exp_correct_bop: forall G bop lxn lyn
    (IN: Ini lxn G)
    (IN: Ini lyn G),
  tm_bsim G (fun n => map enc (tr_exp (IEbop bop lxn lyn) n (fvars G) (tvars G) (kvars G)))
            (IEbop bop lxn lyn).
Proof.
  ii; subst; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  hexploit (values_for_varnat RSV RDV IN0); i; des.
  hexploit (values_for_varnat RSV RDV IN); i; des.

  revert CR; tm_prepare'.

  destruct (LEM (exists n n0, vb = IVnat n /\ vb0 = IVnat n0)); last first.
  { 
    pfold; econs; i.
    destruct cfb; rrdes XTRb; clarify.
    invc CNF; des; subst; noemp.
    rrdes MB; s in Ec; clarify.
    econs 1.
    - instantiate (1:=(None, (envb, IElet (IEbop bop lxn lyn) (inl x) kb))).
      econs 1; eapply IRerr; vauto; i; invc FH.
      s in EVAL; rewrite INb INb0 in EVAL; destruct vb, vb0; clarify; eauto 10.
    - eauto using esim_err.
  }
  des; subst.

  eapply pesim_stepb_obsolete.
  { i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits.
    - econs 1; inst_pairs; eapply IRok.
      eapply IRlet; s; rewrite INb INb0; eauto.
    - inst_pairs; s; noemp; rr; esplits; eauto.
  }

  eapply pesim_getValNat; eauto; [| | |by eapply wfo_le_n_add|].
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr; pmplus_eq'.
  }
  { eapply reprlist_mon; eauto.
    i; eapply pmread_plus; eauto.
    pmplus_eq'.
  }
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr map_app pmcells_app SORD; pmplus_eq'.
  }
  clear INa0.

  eapply pesim_getValNat; eauto; [| |omega| |by eapply wfo_le_n_add|].
  { instantiate (1 := length (getValNat lxn n (map fst senva) (tvars G) r_ret) + n).
    i; eapply pmread_plus; eauto.
    rewrite Eladdr SORD.
    replace (_ - _) with (mpc - n); last first.
    - clear; set (x := length _); omega.
    - pmplus_eq'.
  }
  { tmext_simpl; eapply reprlist_mon; eauto.
    i; eapply pmread_plus; eauto.
    pmplus_eq'.
  }
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr !map_app !pmcells_app !map_length SORD; pmplus_eq'.
  }

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0.
  - clear SV HV FIN; rename mrf into mrf0.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr]; eauto.
      eapply pmread_plus; eauto.
      rewrite Eladdr !map_app !pmcells_app !map_length; s.
      rewrite ->?pmcells_cons, SORD; pmplus_eq'.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr !map_app !pmcells_app !map_length; s.
        rewrite ->?pmcells_cons, SORD; pmplus_eq'.
      - replace (mrf0 r_arg) with (add_offset (mrf r_arg) o_zero)
        ; last (by s; unfold mrf; tmext_simpl; auto); eauto.
    }
    { depmcplus0; eauto using pmvalid_plus_cell, pmfinite_plus_cell. }

    eapply dance_O; instantiate (1 := fun _ _ => _); s; eauto.

  - i; rrdes XTRb RLb0; clarify.
    s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.

      setcrf mrf'.
      set (s' := (mrf', P0, s, s3, (fst s4, {mrf r_arg |-> bosem bop va0 va} ## snd s4))).
      assert (VLD': pmvalid ({mrf r_arg |-> bosem bop va0 va} ## snd s4)).
      { decomp_pmvalid HV; auto. }
      assert (EXT': extpub (wlift mylw) s' (mrf, P0, s, s3, s4)).
      { clear -HV VLD'. unfold s', mrf'.
        s; splits; tmext_simpl; auto.
        i; eapply pmread_plus; eauto.
        rewrite ->pmplus_comm; eauto.
      }
      exists s'; split; [apply extpub_in_ext; exact EXT'|].

      rewrite tgt_csplit; apply pesim_release; last first; try by rr; s; eauto.
      { unfold s', mrf'.
        rr; esplits; last by eauto.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto.
            instantiate (1 := mkconf _ _ _ _).
            rrdes CPa; rr; esplits; s; tmext_simpl; eauto.
          + s; eapply tm_refconfrel_mon; s; unfold extend
            ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
            ; s; i; eapply gk_mon; eauto
            ; rrdes EXT'; rr; eauto using extpub_in_ext.
        - rr; esplits; eauto.
        - s; noemp; lhs rewrite -pmplus_swap -!pmplus_assoc; eauto.
      }

      simpl tr_exp in KSIM; rewrite !map_length !app_length in KSIM; simpl length in KSIM.
      rewrite plus_assoc -(plus_comm 2) in KSIM; simpl plus in KSIM.
      rewrite plus_comm plus_assoc SORD in KSIM.
      inst_pairs; eapply KSIM.
      * exact EXT'.
      * unfold s', mrf'; s; tmext_simpl; auto.
      * unfold s', mrf'; s; tmext_simpl; eapply pmread_plus; eauto.
      * econs; eapply vclgen_nat; s; eauto.
    + apply tm_vclosDb in VCL; apply tm_vclosDb in VCL0.
      ss; subst; noemp; try rewrite plus_comm; rr; esplits; s; eauto.
Qed.



Lemma tr_exp_correct_refeq: forall G lx ly
    (IN: In (inl lx) G)
    (IN: In (inl ly) G),
  tm_bsim G (fun n => map enc (tr_exp (IErefeq lx ly) n (fvars G) (tvars G) (kvars G)))
            (IErefeq lx ly).
Proof.
  ii; subst; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  hexploit (values_for_var RSV RDV IN0); i; des.
  hexploit (values_for_var RSV RDV IN); i; des.

  revert CR; tm_prepare'.

  destruct (LEM (exists l l0, vb = IVloc l /\ vb0 = IVloc l0)); last first.
  { 
    pfold; econs; i.
    destruct cfb; rrdes XTRb; clarify.
    invc CNF; des; subst; noemp.
    rrdes MB; s in Ec; clarify.
    econs 1.
    - instantiate (1:=(None, (envb, IElet (IErefeq lx ly) (inl x) kb))).
      econs 1; eapply IRerr; vauto; i; invc FH.
      s in EVAL; rewrite INb INb0 in EVAL; destruct vb, vb0; clarify; eauto 10.
    - eauto using esim_err.
  }
  des; subst.

  eapply pesim_stepb_obsolete.
  { i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits.
    - econs 1; inst_pairs; eapply IRok.
      eapply IRlet; s; rewrite INb INb0; eauto.
    - inst_pairs; s; noemp; rr; esplits; eauto.
  }

  eapply pesim_getVal; eauto; [| | |by eapply wfo_le_n_add|].
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr; pmplus_eq'.
  }
  { eapply reprlist_mon; eauto.
    i; eapply pmread_plus; eauto.
    pmplus_eq'.
  }
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr map_app pmcells_app SORD; pmplus_eq'.
  }
  clear INa0.

  eapply pesim_getVal; eauto; [| |omega| |by eapply wfo_le_n_add|].
  { instantiate (1 := length (getVal lx n (map fst senva) (tvars G) r_ret) + n).
    i; eapply pmread_plus; eauto.
    rewrite Eladdr SORD.
    replace (_ - _) with (mpc - n); last first.
    - clear; set (x := length _); omega.
    - pmplus_eq'.
  }
  { tmext_simpl; eapply reprlist_mon; eauto.
    i; eapply pmread_plus; eauto.
    pmplus_eq'.
  }
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr !map_app !pmcells_app !map_length SORD; pmplus_eq'.
  }

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0.
  - clear SV HV FIN; rename mrf into mrf0.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr]; eauto.
      eapply pmread_plus; eauto.
      rewrite Eladdr !map_app !pmcells_app !map_length; s.
      rewrite ->?pmcells_cons, SORD; pmplus_eq'.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr !map_app !pmcells_app !map_length; s.
        rewrite ->?pmcells_cons, SORD; pmplus_eq'.
      - replace (mrf0 r_arg) with (add_offset (mrf r_arg) o_zero)
        ; last (by s; unfold mrf; tmext_simpl; auto); eauto.
    }
    { depmcplus0; eauto using pmvalid_plus_cell, pmfinite_plus_cell. }

    eapply dance_O; instantiate (1 := fun _ _ => _); s; eauto.

  - i; rrdes XTRb RLb0; clarify.
    s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.

      setcrf mrf'.
      set (s' := (mrf', P0, s, s3, (fst s4, {mrf r_arg |-> bosem Equal va0 va} ## snd s4))).
      assert (VLD': pmvalid ({mrf r_arg |-> bosem Equal va0 va} ## snd s4)).
      { decomp_pmvalid HV; auto. }
      assert (EXT': extpub (wlift mylw) s' (mrf, P0, s, s3, s4)).
      { clear -HV VLD'. unfold s', mrf'.
        s; splits; tmext_simpl; auto.
        i; eapply pmread_plus; eauto.
        rewrite ->pmplus_comm; eauto.
      }
      exists s'; split; [apply extpub_in_ext; exact EXT'|].

      rewrite tgt_csplit; apply pesim_release; last first; try by rr; s; eauto.
      { unfold s', mrf'.
        rr; esplits; last by eauto.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto.
            instantiate (1 := mkconf _ _ _ _).
            rrdes CPa; rr; esplits; s; tmext_simpl; eauto.
          + s; eapply tm_refconfrel_mon; s; unfold extend
            ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
            ; s; i; eapply gk_mon; eauto
            ; rrdes EXT'; rr; eauto using extpub_in_ext.
        - rr; esplits; eauto.
        - s; noemp; lhs rewrite -pmplus_swap -!pmplus_assoc; eauto.
      }

      simpl tr_exp in KSIM; rewrite !map_length !app_length in KSIM; simpl length in KSIM.
      rewrite plus_assoc -(plus_comm 2) in KSIM; simpl plus in KSIM.
      rewrite plus_comm plus_assoc SORD in KSIM.
      inst_pairs; eapply KSIM.
      * exact EXT'.
      * unfold s', mrf'; s; tmext_simpl; auto.
      * unfold s', mrf'; s; tmext_simpl; eapply pmread_plus; eauto.
      * econs; eapply vclgen_nat; s; eauto.
    + apply tm_vclosDb in VCL; apply tm_vclosDb in VCL0.
      ss; subst; noemp; try rewrite plus_comm; rr; esplits; s; eauto.
      invc VCL VCL0. des. move: (refst_bij REG REG0) => [H1 H2].
      inversion Eb; inversion Eb0; ss.
      case (eq_nat_dec x1 x0); case (eq_nat_dec va0 va); auto; ii.
      * by destruct (n0 (eq_sym (H2 (eq_sym e)))).
      * by destruct (n0 (eq_sym (H1 (eq_sym e)))).
Qed.




Lemma tr_exp_correct_gen: forall G k e
    (GT: is_good_thunk (IEgen k e))
    (IHe: tm_psim (inr k :: purge G)
            (fun n => map enc (tr_exp e n (fvars G) (tvars G) [k]))
            e),
  tm_bsim G (fun n => map enc (tr_exp (IEgen k e) n (fvars G) (tvars G) (kvars G)))
            (IEgen k e).
Proof.
  ii; subst; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  specialize (IHe j); rrdes IHe.
  revert CR; tm_prepare'.

  eapply pesim_stepb_obsolete.
  { i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits.
    - econs 1; inst_pairs; eapply IRok.
      eapply IRlet; s; eauto.
    - inst_pairs; s; noemp; rr; esplits; eauto.
  }

  eapply pesim_dance; first (by eauto); clear RLa0.

  - clear SV HV FIN.

    dance_init.
    { eapply Sok; [by vauto|eapply Sld]; [by read_code Eladdr|by eauto]. }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Salc]; [by read_code Eladdr|by eauto]. }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [by vauto|eapply Slpc]; read_code Eladdr. }
    { depmcplus0. }
    tmext_uniq.

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr]; [by read_code Eladdr| | |]; eauto. }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
      rewrite pmcells_cons -pmplus_assoc.
      replace newloc with (add_offset (mrf r_clo) o_zero)
      ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; pmcells_fold; fold newloc; tmext_simpl.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
      rewrite 2!pmcells_cons -?pmplus_assoc (pmplus_swap {_|->_}).
      replace (S newloc) with (add_offset (mrf r_clo) o_one)
      ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; tmext_simpl; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
      rewrite (pmplus_swap {_|=>_}).
      replace (mrf0 r_arg)  with (add_offset (mrf r_arg) o_zero)
      ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; tmext_simpl; rewrite (pmplus_swap {_|=>_}) in HV0, FIN0.
      depmcplus0; eauto using pmvalid_plus_cell, pmfinite_plus_cell.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr]; [by read_code Eladdr| | |]; eauto. }
    { depmcplus0. }
    tmext_uniq.

    dance_init.
    { eapply Sok; [by vauto|eapply Sjmp]; [by read_code Eladdr|by eauto]. }
    { depmcplus0. }

    eapply dance_O; instantiate (1 := fun _ cf _ _ _ _ _ _ _ => depmach1 cf).
    s; tmext_simpl; eauto.

  - i; rrdes XTRb RLb0; clarify.
    s in Ec; clarify.
    esplits; first 1 last.
    { right; split; eauto using wfo_le_n_add. }
    { inst_pairs; rr; esplits; s; noemp; eauto. }
    
    i; frames_dep 1 RLa XTRa cfas; frames_nodep RLa cfas.
    destruct ea'; clear SV HV FIN.
    destruct_tgt_real RLa.
    setnewloc.
    simpl lang_tgt.chp in *.
    setcrf mrf'.
    set (s' := (mrf', extend P0 (ts_fun (S (S (mpc + 7)))) newloc, s, s3,
                (fst s4, {mrf r_arg |-> newloc} ## {S newloc |-> mrf r_env} ## snd s4))).

    assert (VLD': pmvalid ({mrf r_arg |-> newloc} ## {S newloc |-> mrf r_env} ## snd s4)).
    { clear -HV.
      rewrite pmcells_cons -pmplus_assoc (pmplus_swap {newloc |-> _}) -pmcells_sng in HV.
      rewrite <-(pmplus_comm (snd s4)), <-2(pmplus_swap (snd s4)) in HV.
      do 2 eapply pmvalid_plus_l; rewrite <-3pmplus_assoc; eauto.
    }

    assert (EXT': extpub (wlift mylw) s' (mrf, P0, s, s3, s4)).
    { unfold s', mrf'; clear -HV VLD'.
      s; splits; tmext_simpl; auto; [by i; right; eauto|].
      i; eapply pmread_plus; eauto; pmplus_eq'.
    }

    exists s'; esplits; [by eauto|].
    rewrite tgt_csplit mid_csplit.
    eapply pesim_release; last first; try by rr; s; eauto.
    { unfold s'.
      rr; esplits.
      - rr; esplits; eauto.
        + s; rr; esplits; eauto. clean.
          instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest;
                               chp := {newloc |-> S (S (mpc + 7))} ## hp |}).
          s; rrdes CPa; s; s in HP.
          unfold mrf'; rr; splits; s; tmext_simpl; auto.
          unfold extend; i; des.
          * rewrite -IN -IN0; s; split; eauto.
            rewrite ->!pmcells_cons, <-!pmplus_assoc in HV.
            decomp_pmvalid HV; eauto.
          * eapply repr_mon; [s; i|by eauto].
            eapply pmread_plus; try rewrite ->pmplus_comm; eauto.
            rewrite ->!pmcells_cons, <-!pmplus_assoc in HV.
            decomp_pmvalid HV; rewrite pmplus_comm; eauto.
        + s; eapply tm_refconfrel_mon; s; unfold extend
          ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
          ; s; i; eapply gk_mon; eauto
          ; rrdes EXT'; rr; eauto using extpub_in_ext.
      - rr; esplits; eauto.
      - s; noemp; f_equal.
        rewrite <-!pmplus_assoc, <-2(pmplus_swap {mrf r_arg|->_}).
        rewrite <-3(pmplus_swap {S newloc|->_}); rhs rewrite pmplus_swap.
        pmcells_fold; eauto.
      - noemp; auto.
    }

    replace (S _) with (length (map enc (tr_exp (IEgen k e) n (fvars G) (tvars G) (kvars G))) + mpc); last first.
    { rewrite map_length; s; omega. }

    eapply KSIM; eauto.
    { unfold s', mrf'; s; tmext_simpl; auto. }
    { unfold s', mrf'; s; tmext_simpl; eauto using pmread_plus. }

    econs; apply vclgen_thunk, vclf_gen.
    { unfold s'; econs; rr; eauto. }
    { rr; eauto. }

    eapply gk_lsim; eauto.
    econs. eapply gt_gen.
    { unfold s'; econs; rr; eauto. }
    { s. econs. i. apply GT. inv FH. }

    i; invc STK; [by invc APPb RF CB VR|].
    eapply pesim_gknow; i.
    invc INSTa INSTb.
    invc RF CB VR. rrdes CA.
    unfold s' in *. clear s'. unfold mrf' in *; clear mrf'. subst.
    destruct s'0 as [[[[]]]]; clean.
    ren PC. ren SV. ren HV. ren FIN.
    destruct s7.
    tm_prepare.
    setnewloc.
    rewrite <-H0 in *.

    assert (Epc: mpc0 = 9 + mpc).
    { rrdes EXT0 EXT0 EXT0 EXT0.
      hexploit tgt_vquery_fun_uniq; [by eauto|by eauto| |].
      - eapply EXT0; s; rr; eauto.
      - rewrite plus_comm; eauto.
    }
    subst mpc0. clear CA PC0.

    clear CPa CPb SV0 RLa0 freest.               
    hexploit (tgt_stack CPa0 2); i; des; s in FST; subst freest0.
    
    des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
    rewrite <-!pmplus_assoc in *; clear NST.

    hexploit (@pmplus_from_read _ _ _ (snd s4) h1).
    { i; rrdes EXT0; apply EXT1; s.
      eapply pmread_plus; eauto; pmplus_eq'.
    }
    { split_pmvalid VLD'; eauto. }
    i; des; subst h1.

    eapply pesim_dance; first (by eauto); clear RLa1.
    + clear SV HV FIN HV0 FIN0 KSIM.

      dance_init. 
      { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
        rewrite pmcells_cons -pmplus_assoc.
        replace (mrf1 r_sp) with (add_offset (mrf1 r_sp) o_zero)
        ; last (by s; tmext_simpl; eauto); eauto.
      }
      { s; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
        rewrite 2!pmcells_cons -2!pmplus_assoc pmplus_swap.
        replace (S (mrf0 r_sp)) with (add_offset (mrf0 r_sp) o_one)
        ; last (by s; tmext_simpl; eauto); eauto.
      }
      { s; pmcells_order; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      dance_init.
      { eapply Sok; [by vauto|eapply Sopr]; [by read_code Eladdr| | |]; eauto. }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [by vauto|eapply Sld]; [by read_code Eladdr|econs].
        rewrite (pmplus_assoc (snd s4)). set (tmp := (snd s4 ## _)).
        tmext_simpl; rrdes EXT0; rewrite -H0.
        eapply pmread_plus; first (eapply EXT1).
        - s; eapply pmread_plus; eauto; pmplus_eq'.
        - s. fold tmp. pmplus_eq'.
        - unfold tmp. rewrite -pmplus_assoc; eauto.
      }
      { depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells. }

      eapply dance_O; instantiate (1 := fun _ _ _ _ => _).
      s; eauto.

    + i; rrdes XTRb RLb0; clarify.
      esplits; first 1 last.
      { right; split; eauto using wfo_le_n_add. }
      { eauto. }
    
      i; frames_nodep RLa cfas.
      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.

      setcrf mrf'.
      set (s' := (mrf', P1, s5, s6, ({mrf0 r_sp |=> [mrf0 r_ret; mrf0 r_env]} ## s7,
                  snd s4 ## h''))).
      assert (EXT'': ext (wlift mylw) s' (mrf0, P1, s5, s6, (s7, snd s4 ## h''))).
      { s; splits; tmext_simpl; auto. }
      exists s'; esplits; [exact EXT''|].

      rewrite tgt_csplit mid_csplit; eapply pesim_release; last first; try by rr; s; eauto.
      { unfold s'; rr; esplits.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto. clean.
            instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest; chp := hp0 |}).
            rrdes CPa0; s in HP.
            unfold mrf'; rr; splits; s; tmext_simpl; auto.
            s in ST.
            assert (XX := tgt_stack_transfer _ _ _ ST). s in XX. eauto.
          + s; eapply tm_refconfrel_mon; s; unfold extend
            ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
            ; s; i; eapply gk_mon; eauto
            ; rrdes EXT'; try split; rr; eauto.
      - rr; esplits; eauto.
      - s; noemp; f_equal.
        + rewrite pmplus_swap; eauto.
        + rewrite (pmplus_assoc hp0); eauto.
      - noemp; auto.
      }

      eapply IHe; eauto.
      * i; s in IN; des; vauto.
        apply purge_inD in IN; hexploit RSV; eauto; i; des.
        esplits; eauto. eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * s; rewrite fvars_purge; s; eauto.
      * instantiate (1 := denva).
        i; apply in_consD in IN; des; vauto.
        apply purge_inD in IN; hexploit RDV; eauto; i; des.
        esplits; eauto.
        eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * unfold s', mrf'; s; tmext_simpl; eauto.
        eapply reprlist_mon; eauto.
        decomp_pmvalid HV.
        i; eapply pmread_plus; eauto.
      * s; rewrite tvars_purge; rewrite DORD; auto.
      * unfold s'; s; eauto.
      * s; rewrite kvars_purge; s; eauto.
      * unfold s', mrf'; s; tmext_simpl.
        rewrite kvars_purge; s; omega.
      * unfold s', mrf'; tmext_simpl; s.
        econs.
        { s; eauto. }
        { rewrite kvars_purge; s.
          econs. s. tmext_simpl. omega.
        }
        { s; tmext_simpl.
          rewrite plus_comm plus_minus_eq; last omega; s.
          decomp_pmvalid SV.
          eapply pmread_plus; eauto.
          rewrite pmcells_cons; pmplus_eq'.
        }
        { s; tmext_simpl.
          rewrite plus_comm; s; rewrite plus_comm plus_minus_eq; last omega; s.
          decomp_pmvalid SV.
          eapply pmread_plus; eauto.
          rewrite 2!pmcells_cons; pmplus_eq'.
        }
        { rewrite list_assoc_cons_eq; eauto. }

        i. invc RETa.
        invc RETb.
        invc RK.                        
        eapply tm_pesim_return; eauto.
        {
          i; subst; rrdes EXT1 EXT1 EXT1; split; [repeat split|]; s; eauto.
          - ss; rewrite SP; tmext_simpl.
            rewrite plus_comm plus_minus_eq; s; auto.
          - split.
            + ii; s in EXT0; unnw; hexploit EXT0; eauto.
            + ss; rewrite SP in LST; tmext_simplH LST.
              rewrite plus_comm plus_minus_eq in LST; last auto; s in LST.
              rewrite (pmplus_comm (fst _)) in LST; decomp_pmvalid SV.
              apply pmcells_uniq in LST; des; eauto.
        }
      * 
        unfold s'; s; econs; symmetry.
        rewrite Eladdr; s; rewrite !pmcells_cons SORD.
        rewrite -!(pmplus_swap {mpc - n |=> _}).
        pmplus_eq'.
Qed.



Lemma tr_exp_correct_fix: forall G f x k e
    (GT: is_good_thunk (IEfix f x k e))
    (IHe: tm_psim (inr k :: inl (inr x) :: inl (inr f) :: purge G)
            (fun n => map enc (tr_exp e n (fvars G) (x :: f :: tvars G) [k]))
            e),
  tm_bsim G (fun n => map enc (tr_exp (IEfix f x k e) n (fvars G) (tvars G) (kvars G)))
            (IEfix f x k e).
Proof.
  ii; subst; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  specialize (IHe j); rrdes IHe.
  revert CR; tm_prepare'.

  eapply pesim_stepb_obsolete.
  { i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits.
    - econs 1; inst_pairs; eapply IRok.
      eapply IRlet; s; eauto.
    - inst_pairs; s; noemp; rr; esplits; eauto.
  }

  
  

  eapply pesim_dance; first (by eauto); clear RLa0.
  
  

  - clear SV HV FIN.

    dance_init.
    { eapply Sok; [by vauto|eapply Sld]; [by read_code Eladdr|by eauto]. }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Salc]; [by read_code Eladdr|by eauto]. }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [by vauto|eapply Slpc]; read_code Eladdr. }
    { depmcplus0. }
    tmext_uniq.

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr]; [by read_code Eladdr| | |]; eauto. }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
      rewrite pmcells_cons -pmplus_assoc.
      replace newloc with (add_offset (mrf r_clo) o_zero)
      ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; pmcells_fold; fold newloc; tmext_simpl.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
      rewrite 2!pmcells_cons -?pmplus_assoc (pmplus_swap {_|->_}).
      replace (S newloc) with (add_offset (mrf r_clo) o_one)
      ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; tmext_simpl; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
      rewrite (pmplus_swap {_|=>_}).
      replace (mrf0 r_arg)  with (add_offset (mrf r_arg) o_zero)
      ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; tmext_simpl; rewrite (pmplus_swap {_|=>_}) in HV0, FIN0.
      depmcplus0; eauto using pmvalid_plus_cell, pmfinite_plus_cell.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr]; [by read_code Eladdr| | |]; eauto. }
    { depmcplus0. }
    tmext_uniq.

    dance_init.
    { eapply Sok; [by vauto|eapply Sjmp]; [by read_code Eladdr|by eauto]. }
    { depmcplus0. }

    eapply dance_O; instantiate (1 := fun _ cf _ _ _ _ _ _ _ => depmach1 cf).
    s; tmext_simpl; eauto.

  - i; rrdes XTRb RLb0; clarify.
    s in Ec; clarify.
    esplits; first 1 last.
    { right; split; eauto using wfo_le_n_add. }
    { inst_pairs; rr; esplits; s; noemp; eauto. }
    
    i; frames_dep 1 RLa XTRa cfas; frames_nodep RLa cfas.
    destruct ea'; clear SV HV FIN.
    destruct_tgt_real RLa.
    setnewloc.
    simpl lang_tgt.chp in *.
    setcrf mrf'.
    set (s' := (mrf', extend P0 (ts_fun (S (S (mpc + 7)))) newloc, s, s3,
                (fst s4, {mrf r_arg |-> newloc} ## {S newloc |-> mrf r_env} ## snd s4))).

    assert (VLD': pmvalid ({mrf r_arg |-> newloc} ## {S newloc |-> mrf r_env} ## snd s4)).
    { clear -HV.
      rewrite pmcells_cons -pmplus_assoc (pmplus_swap {newloc |-> _}) -pmcells_sng in HV.
      rewrite <-(pmplus_comm (snd s4)), <-2(pmplus_swap (snd s4)) in HV.
      do 2 eapply pmvalid_plus_l; rewrite <-3pmplus_assoc; eauto.
    }

    assert (EXT': extpub (wlift mylw) s' (mrf, P0, s, s3, s4)).
    { unfold s', mrf'; clear -HV VLD'.
      s; splits; tmext_simpl; auto; [by i; right; eauto|].
      i; eapply pmread_plus; eauto; pmplus_eq'.
    }

    exists s'; esplits; [by eauto|].
    rewrite tgt_csplit mid_csplit.
    eapply pesim_release; last first; try by rr; s; eauto.
    { unfold s'.
      rr; esplits.
      - rr; esplits; eauto.
        + s; rr; esplits; eauto. clean.
          instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest;
                               chp := {newloc |-> S (S (mpc + 7))} ## hp |}).
          s; rrdes CPa; s; s in HP.
          unfold mrf'; rr; splits; s; tmext_simpl; auto.
          unfold extend; i; des.
          * rewrite -IN -IN0; s; split; eauto.
            rewrite ->!pmcells_cons, <-!pmplus_assoc in HV.
            decomp_pmvalid HV; eauto.
          * eapply repr_mon; [s; i|by eauto].
            eapply pmread_plus; try rewrite ->pmplus_comm; eauto.
            rewrite ->!pmcells_cons, <-!pmplus_assoc in HV.
            decomp_pmvalid HV; rewrite pmplus_comm; eauto.
        + s; eapply tm_refconfrel_mon; s; unfold extend
          ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
          ; s; i; eapply gk_mon; eauto
          ; rrdes EXT'; rr; eauto using extpub_in_ext.
      - rr; esplits; eauto.
      - s; noemp; f_equal.
        rewrite <-!pmplus_assoc, <-2(pmplus_swap {mrf r_arg|->_}).
        rewrite <-3(pmplus_swap {S newloc|->_}); rhs rewrite pmplus_swap.
        pmcells_fold; eauto.
      - noemp; auto.
    }

    replace (S _) with (length (map enc (tr_exp (IEfix f x k e) n (fvars G) (tvars G) (kvars G))) + mpc)
    ; last by rewrite map_length; s; omega.

    eapply KSIM; eauto.
    { unfold s', mrf'; s; tmext_simpl; auto. }
    { unfold s', mrf'; s; tmext_simpl; eauto using pmread_plus. }

    econs; apply vclgen_thunk, vclf_fix.
    { unfold s'; econs; rr; eauto. }
    { rr; eauto. }

    eapply gk_lsim; eauto.
    evar (i0: wfo_set (wlift mylw)).             

    econs. econs.
    { unfold s'; econs; rr; eauto. }
    { s. econs. i. apply GT. inv FH. }

    instantiate (1:=i0).
    pcofix CIH. unfold i0 in *. clear i0.
    i; invc STK; [|by invc INSTb RF CB VR].
    eapply pesim_gknow; i; eapply pesim_gknow_; i.
    invc APPa APPb.
    invc RF CB VR. rrdes CA.
    unfold s' in *. clear s'. unfold mrf' in *; clear mrf'. subst.
    destruct s'0 as [[[[]]]]; clean.
    ren PC. ren SV. ren HV. ren FIN.
    destruct s7.
    tm_prepare.
    setnewloc.
    rewrite <-H0 in *.

    assert (Epc: mpc0 = 9 + mpc).
    { rrdes EXT0 EXT0 EXT0 EXT0.
      hexploit tgt_vquery_fun_uniq; [by eauto|by eauto| |].
      - eapply EXT0; s; rr; eauto.
      - rewrite plus_comm; eauto.
    }
    subst mpc0. clear CA PC0.

    clear CPa CPb SV0 RLa0 freest.               
    hexploit (tgt_stack CPa0 2); i; des; s in FST; subst freest0.
    
    des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
    rewrite <-!pmplus_assoc in *; clear NST.

    hexploit (@pmplus_from_read _ _ _ (snd s4) h1).
    { i; rrdes EXT0; apply EXT1; s.
      eapply pmread_plus; eauto; pmplus_eq'.
    }
    { split_pmvalid VLD'; eauto. }
    i; des; subst h1.

    eapply pesim_dance; first (by eauto); clear RLa1.
    + clear SV HV FIN HV0 FIN0 KSIM.

      dance_init. 
      { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
        rewrite pmcells_cons -pmplus_assoc.
        replace (mrf1 r_sp) with (add_offset (mrf1 r_sp) o_zero)
        ; last (by s; tmext_simpl; eauto); eauto.
      }
      { s; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
        rewrite 2!pmcells_cons -2!pmplus_assoc pmplus_swap.
        replace (S (mrf0 r_sp)) with (add_offset (mrf0 r_sp) o_one)
        ; last (by s; tmext_simpl; eauto); eauto.
      }
      { s; pmcells_order; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      dance_init.
      { eapply Sok; [by vauto|eapply Sopr]; [by read_code Eladdr| | |]; eauto. }
      { instantiate (1 := depmach0).
        eapply tgt_mcplus; first by s; eauto.
        rr; esplits; s; try pmplus_eq; eauto.
      }

      dance_init.
      { eapply Sok; [by vauto|eapply Sld]; [by read_code Eladdr|by eauto]. }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [by vauto|eapply Salc]; [by read_code Eladdr|by eauto]. }
      { depmcplus1; eauto using alloc_valid. }
      s; (Hdo (fun H => try fold newloc0 in H)); setnewloc. 

      rename mrf into mrf00.    

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
        tmext_simpl; rewrite (pmcells_cons newloc) -pmplus_assoc.
        replace newloc with (add_offset (mrf r_env) o_zero)
        ; last (by unfold mrf; s; tmext_simpl; eauto); eauto.
      }
      { s; tmext_simpl; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }
      fold newloc.              

      dance_init.
      { eapply Sok; [by vauto|eapply Sld]; [by read_code Eladdr|econs].
        rewrite (pmplus_assoc (snd s4)). set (tmp := (snd s4 ## _)).
        unfold mrf; s; tmext_simpl; rrdes EXT0; rewrite -H0.
        eapply pmread_plus; first (eapply EXT1).
        - s; eapply pmread_plus; eauto; pmplus_eq'.
        - s. fold tmp. pmplus_eq'.
        - unfold tmp; rewrite -pmplus_assoc H0; eauto.
      }                         
      { depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
        rewrite (pmcells_cons newloc) -pmplus_assoc.
        rewrite (pmcells_cons (S newloc)) -pmplus_assoc. 
        rewrite (pmplus_swap {_|->_}).
        replace (S newloc) with (add_offset (mrf r_env) o_one)
        ; last (by unfold mrf; s; tmext_simpl; eauto); eauto.
      }
      { s; tmext_simpl; pmcells_order; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      dance_init.
      { eapply Sok; [by vauto|eapply Sld]; [by read_code Eladdr|by eauto]. }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [by vauto|eapply Salc]; [by read_code Eladdr|by eauto]. }
      { depmcplus1; eauto using alloc_valid. }
      Hdo (fun H => try fold newloc0 in H).
      s; setnewloc.

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
        rewrite (pmcells_cons newloc1) -pmplus_assoc.
        replace newloc1 with (add_offset (mrf r_clo) o_zero)
        ; last (by unfold mrf; s; tmext_simpl; eauto); eauto.
      }
      { s; tmext_simpl; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }
      fold newloc1; setnewloc.

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
        rewrite (pmcells_cons newloc1) -pmplus_assoc.
        rewrite (pmcells_cons (S newloc1)) -pmplus_assoc.
        rewrite (pmplus_swap {_|->_}).
        replace (S newloc1) with (add_offset (mrf r_clo) o_one)
        ; last (by unfold mrf; s; tmext_simpl; eauto); eauto.
      }
      { s; tmext_simpl; pmcells_order; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      dance_init.
      { eapply Sok; [by vauto|eapply Sld]; [by read_code Eladdr|by eauto]. }
      { depmcplus0. }
      tmext_uniq.

      eapply dance_O; instantiate (1 := fun _ _ _ _ cf _ _ _ _ cf' _ _ _ => depmach2 cf cf').
      s; eauto.

    + i; rrdes XTRb RLb0; clarify.
      esplits; first 1 last.
      { right; split; eauto using wfo_le_n_add. }
      { eauto. }
    
      i; frames_dep 4 RLa XTRa cfas.
      frames_dep 4 RLa XTRa cfas.
      frames_nodep RLa cfas.

      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.
      Hdo (fun H => try fold newloc0 in H).
      setnewloc.

      match type of RLa0 with tgt_real {| cpc := _; crf := Just ?rf; cst := _; chp := _ |} _ =>
        set (mrf' := rf)
      end.
      set (s' := (mrf', P1, s5, s6, ({mrf0 r_sp |=> [mrf0 r_ret; mrf0 r_env]} ## s7,
                  {newloc1 |=> [mrf0 r_arg; newloc2]} ##
                  {newloc2 |=> [mrf0 r_clo; mrf r_env]} ##
                  snd s4 ## h''))).

      assert (VLD'': pmvalid ({newloc1 |=> [mrf0 r_arg; newloc2]} ##
                              {newloc2 |=> [mrf0 r_clo; mrf r_env]} ##
                              snd s4 ## h'')).
      { clear -HV.
        rewrite -!(pmplus_swap hp0) -!(pmplus_swap refha0) in HV.
        eapply pmvalid_plusD, pmvalid_plusD; eauto. 
      }

      assert (EXT'': ext (wlift mylw) s' (mrf0, P1, s5, s6, (s7, snd s4 ## h''))).
      { clear -VLD''.
        s; splits; tmext_simpl; auto.
        i; eapply pmread_plus; eauto.
        set (tmp := snd s4 ## _); pmplus_eq'.
      }

      exists s'; esplits; [exact EXT''|].
      rewrite tgt_csplit mid_csplit.
      eapply pesim_release; last first; try by rr; s; eauto.
      { unfold s'.
        rr; esplits.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto. clean.
            instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest; chp := hp0 |}).
            rrdes CPa0; s in HP.
            unfold mrf'; rr; splits; s; tmext_simpl; auto.
            s in ST.
            assert (XX := tgt_stack_transfer _ _ _ ST). s in XX. eauto.
          + s; eapply tm_refconfrel_mon; s; unfold extend
            ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
            ; s; i; eapply gk_mon; eauto
            ; rrdes EXT''; try split; rr; eauto.
      - rr; esplits; eauto.
      - s; noemp; f_equal.
        + rewrite pmplus_swap; eauto.
        + rewrite (pmplus_assoc hp0); set (tmp := hp0 ## _).
          rewrite -!(pmplus_swap tmp); eauto.
      - noemp; auto.
      }

      eapply paco9_mon; [eapply IHe|by eauto]; eauto.
      * i; s in IN; des; vauto.
        apply purge_inD in IN; hexploit RSV; eauto; i; des.
        esplits; eauto. eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * s; rewrite fvars_purge; s; eauto.
      * instantiate (1 := (bx, mrf0 r_arg) :: (bf, mrf0 r_clo) :: denva).
        i; apply in_consD in IN; des; vauto.
        apply in_consD in IN; des.
        { clarify.
          rewrite list_assoc_cons_eq.
          rewrite list_assoc_cons_neq; [|by vauto].
          rewrite list_assoc_cons_eq.
          esplits; eauto.
          eapply vclos_mon; eauto using ext_trans.
          i; eapply gk_mon; eauto using ext_trans.
        }
        rewrite list_assoc_cons_neq; [|by ii; vauto].
        apply in_consD in IN; des.
        { clarify.
          rewrite list_assoc_cons_eq.
          rewrite list_assoc_cons_neq; [|by ii; vauto].
          rewrite list_assoc_cons_neq; [|by ii; vauto].
          rewrite list_assoc_cons_eq.
          esplits; eauto.
          econs.                
          apply vclgen_thunk, vclf_fix.
          { rewrite -H0. unfold s'; econs; rr; eauto.
            rrdes EXT0 EXT0 EXT0 EXT0. eapply EXT0.
            s. left; eauto.
          }
          { rr; eauto. }

          eapply gk_lsim; eauto.
          econs. econs.
          { unfold s'; econs; rr; eauto. }
          { s. econs. i. apply GT. inv FH. }
          
          i; eapply gres_incl_id. right.
          eapply CIH; eauto using ext_trans.
          rewrite -H0 in STK. eauto.
        }
        rewrite list_assoc_cons_neq; [|by ii; vauto].
        rewrite list_assoc_cons_neq; [|by ii; vauto].
        rewrite list_assoc_cons_neq; [|by ii; vauto].
        rewrite list_assoc_cons_neq; [|by ii; vauto].
        apply purge_inD in IN; hexploit RDV; eauto; i; des.
        esplits; eauto.
        eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * unfold s', mrf'; s; tmext_simpl; esplits
        ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
        eapply reprlist_mon; eauto.
        i; eapply pmread_plus; eauto; pmplus_eq'.
      * s; rewrite tvars_purge; rewrite DORD; auto.
      * unfold s'; s; eauto.
      * s; rewrite kvars_purge; s; eauto.
      * unfold s', mrf'; s; tmext_simpl.
        rewrite kvars_purge; s; omega.
      * unfold s', mrf'; tmext_simpl; s.
        econs.
        { s; eauto. }
        { rewrite kvars_purge; s.
          econs. s. tmext_simpl. omega.
        }
        { s; tmext_simpl.
          rewrite plus_comm plus_minus_eq; last omega; s.
          decomp_pmvalid SV.
          eapply pmread_plus; eauto.
          rewrite pmcells_cons; pmplus_eq'.
        }
        { s; tmext_simpl.
          rewrite plus_comm; s; rewrite plus_comm plus_minus_eq; last omega; s.
          decomp_pmvalid SV.
          eapply pmread_plus; eauto.
          rewrite 2!pmcells_cons; pmplus_eq'.
        }
        { rewrite list_assoc_cons_eq; eauto. }

        i. invc RETa.
        invc RETb.
        invc RK.                        
        eapply tm_pesim_return; eauto.
        {
          rrdes EXT1 EXT1 EXT1; split; [repeat split|]; s; eauto.
          - ss; rewrite SP; tmext_simpl.
            rewrite plus_comm plus_minus_eq; s; auto.
          - split.
            + ii; s in EXT2; unnw; hexploit EXT2; eauto.
              eapply pmread_plus; eauto.
              set (tmp := _ ## h''); pmplus_eq'.
            + ss; rewrite SP in LST; tmext_simplH LST.
              rewrite plus_comm plus_minus_eq in LST; last auto; s in LST.
              rewrite (pmplus_comm (fst _)) in LST; decomp_pmvalid SV.
              apply pmcells_uniq in LST; des; eauto.
        }
      * unfold s'; s; econs; symmetry.
        rewrite Eladdr; s.
        rhs rewrite -!(pmplus_swap {mpc - n |=> _}).
        
        rhs rewrite -!pmplus_assoc. 
        rhs rewrite -!(pmplus_swap {mpc - n |=> _}).
        pmplus_eq'.
Qed.



(** Final result: any atomic expression---except a continuation---is
    related to its compilation.

    Continuations are excluded because they can refer to older
    continuations, which [tm_bsim] rules out.  This is consistent with
    the compilation of [let] treating continuation definitions as a
    special case.  Accordingly, the proof treats them as a special
    case, see [tr_exp_correct_contvar] in file mid2tgt_correct_ctrl.v.
*)

Lemma tr_exp_correct_atom: forall ae G
    (WF: wfatom_ wfexp G ae)
    (NC: forall x e, ae <> IEcont x e)
    (IH: forall y, y < size ae -> forall e', y = size e' -> forall G, wfexp G e' ->
           tm_psim G (fun n => map enc (tr_exp e' n (fvars G) (tvars G) (kvars G))) e'),
  tm_bsim G (fun n => map enc (tr_exp ae n (fvars G) (tvars G) (kvars G))) ae.
Proof.
  induction ae; i; try invc WF.
  - eapply tr_exp_correct_unit.
  - eapply tr_exp_correct_nat.
  - eapply tr_exp_correct_pair; auto.
  - eapply tr_exp_correct_inl; auto.
  - eapply tr_exp_correct_inr; auto.
  - eapply tr_exp_correct_bop; auto.
  - eapply tr_exp_correct_fst; auto.
  - eapply tr_exp_correct_snd; auto.
  - eapply tr_exp_correct_fix; eauto using wfexp_good_fix.
    hexploit (IH (size ae)); eauto.
    s; rewrite fvars_purge tvars_purge kvars_purge; eauto.
  - eapply tr_exp_correct_gen; eauto using wfexp_good_gen.
    hexploit (IH (size ae)); eauto.
    s; rewrite fvars_purge tvars_purge kvars_purge; eauto.
  - exfalso; eapply NC; auto.
  - eapply tr_exp_correct_refeq; auto.
Qed.
