(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Omega.
Require Import FunctionalExtensionality Classical ClassicalChoice IndefiniteDescription.
Require Import vgtac basis countable myclassical.
Require Import lang_spec lang_common model_common.
Require Import types model model_properties paco9.
               

Set Implicit Arguments.
Global Set Bullet Behavior "Strict Subproofs".


Section esim.

  Variable La Lb: LangSpec.
  Variable gw: world_glob La Lb.
  Variable lwl lwr: world_loca gw.

  Definition le_left_true (esl: esimtyp (wlift lwl)) (es: esimtyp (wlift (wprod lwl lwr))) :=
    forall ord ka kb sig sg sil sl sir sr (EXTP: extpub lwr sr sir) g ty ea eb,
      esl ord ka kb (prod_gka g sr) (sig, sil) (sg, sl) true ty ea eb ->
      es (ord,wfo_zero) ka kb g (sig, (sil, sir)) (sg, (sl, sr)) true ty ea eb.

  Definition le_left_false (esl: esimtyp (wlift lwl)) (es: esimtyp (wlift (wprod lwl lwr))) :=
    forall ord ka kb sig sg sil sl sir sr (EXTP: extpub lwr sr sir) g ty ea eb hra hrb,
      esl ord ka kb (prod_gka g sr) (sig, sil) (sg, sl) false ty ea eb ->
      confrel lwr (prod_gkb g sl) (sg, sr) hra hrb ->
      es (ord,wfo_zero) ka kb g (sig, (sil, sir)) (sg, (sl, sr)) false ty (cplus ea hra) (cplus eb hrb).

  Definition le_right_true (esr: esimtyp (wlift lwr)) (es: esimtyp (wlift (wprod lwl lwr))) :=
    forall ord ka kb sig sg sil sl sir sr (EXTP: extpub lwl sl sil) g ty ea eb,
      esr ord ka kb (prod_gkb g sl) (sig, sir) (sg, sr) true ty ea eb ->
      es (wfo_zero,ord) ka kb g (sig, (sil, sir)) (sg, (sl, sr)) true ty ea eb.

  Definition le_right_false (esr: esimtyp (wlift lwr)) (es: esimtyp (wlift (wprod lwl lwr))) :=
    forall ord ka kb sig sg sil sl sir sr (EXTP: extpub lwl sl sil) g ty ea eb hla hlb,
      esr ord ka kb (prod_gkb g sl) (sig, sir) (sg, sr) false ty ea eb ->
      confrel lwl (prod_gka g sr) (sg, sl) hla hlb ->
      es (wfo_zero,ord) ka kb g (sig, (sil, sir)) (sg, (sl, sr)) false ty (cplus ea hla) (cplus eb hlb).

  
  Lemma prod_gka_mon: forall La Lb (gw: world_glob La Lb) stsa stsb
      g sr s ty va vb (IN: @prod_gka La Lb gw stsa stsb g sr s ty va vb)
      g' (LE: g <g= g')
      sr' (EXT: ext _ sr' sr),
    prod_gka g' sr' s ty va vb.
  Proof.
    i; eapply gmon_mon; eauto.
  Qed.

  Lemma prod_gka_smon: forall La Lb (gw: world_glob La Lb) stsa stsb
      g sr s ty va vb (IN: @prod_gka La Lb gw stsa stsb g sr s ty va vb)
      s' (EXT: ext _ s' s)
      (MON: forall sr', g (fst s, (snd s, sr')) <r= g (fst s', (snd s', sr'))),
    prod_gka g sr s' ty va vb.
  Proof.
    i; eapply gmon_mon; eauto.
  Qed.

  Lemma prod_gkb_mon: forall La Lb (gw: world_glob La Lb) stsa stsb
      g sl s ty va vb (IN: @prod_gkb La Lb gw stsa stsb g sl s ty va vb)
      g' (LE: g <g= g')
      sl' (EXT: ext _ sl' sl),
    prod_gkb g' sl' s ty va vb.
  Proof.
    i; eapply gmon_mon; eauto.
  Qed.                          

  Lemma prod_gkb_smon: forall La Lb (gw: world_glob La Lb) stsa stsb
      g sl s ty va vb (IN: @prod_gkb La Lb gw stsa stsb g sl s ty va vb)
      s' (EXT: ext _ s' s)
      (MON: forall sl', g (fst s, (sl', snd s)) <r= g (fst s', (sl', snd s'))),
    prod_gkb g sl s' ty va vb.
  Proof.
    i; eapply gmon_mon; eauto.
  Qed.


  Lemma lsim_left:
    forall esl es (LE: le_left_true esl es)
           g (G: gknow_ (wlift (wprod lwl lwr)) es g) sg sl sr,
      lsim_ esl (prod_gka g sr) (sg, sl) <r=
      lsim_ es g (sg, (sl, sr)).
  Proof.
    ii; destruct PR; econs; eauto.
    - destruct WF; eauto.
    - i; destruct s' as (?&?&?); rrdes EXT EXT0.
      instantiate (1:= (_,_)).
      eapply LE, CS; s; eauto using prod_gka_mon.
      destruct STK; [econs 1|econs 2]; eauto; [M|M| | |M|M|]; Mdo (ss; eauto).
      + inversion RF; econs; ss; eauto.
      + eapply vclos_unlift in RV; eapply vclos_lift; eauto using vclos_mon.
      + inversion RF; econs; ss; eauto.
  Qed.

  Lemma lsim_right:
    forall esr es (LE: le_right_true esr es)
           g (G: gknow_ (wlift (wprod lwl lwr)) es g) sg sl sr,
      lsim_ esr (prod_gkb g sl) (sg, sr) <r=
      lsim_ es g (sg, (sl, sr)).
  Proof.
    ii; destruct PR; econs; eauto.
    - destruct WF; eauto.
    - i; destruct s' as (?&?&?); rrdes EXT EXT0.
      instantiate (1:= (_,_)).
      eapply LE, CS; s; eauto using prod_gkb_mon.
      destruct STK; [econs 1|econs 2]; eauto; [M|M| | |M|M|]; Mdo (ss; eauto).
      + inversion RF; econs; ss; eauto.
      + eapply vclos_unlift in RV; eapply vclos_lift; eauto using vclos_mon.
      + inversion RF; econs; ss; eauto.
  Qed.


  Lemma ksim_left:
    forall esl es (LE: le_left_true esl es)
           g (G: gknow_ (wlift (wprod lwl lwr)) es g)
           ord kia kib sig spg sg sil spl sl
           sir spr sr (EXTP: extpub lwr spr sir),
      ksim_ esl ord kia kib (prod_gka g sr) (sig, sil) (spg, spl) (sg, sl) <ks=
      ksim_ es (ord,wfo_zero) kia kib g (sig, (sil, sir)) (spg, (spl, spr)) (sg, (sl, sr)).
  Proof.
    ii; inversion PR; subst; econs; eauto.
    i; destruct s' as (?&?&?); rrdes EXTP0 EXTP1 EXT EXT0.
    eapply LE, KR; s; auto; [M| | |M|M]; Mdo (ss; eauto using extpub_trans).
    - eauto using prod_gka_mon.
    - eapply vclos_unlift in RV; eapply vclos_lift; eauto using vclos_mon.
  Qed.

  Lemma ksim_right:
    forall esr es (LE: le_right_true esr es)
           g (G: gknow_ (wlift (wprod lwl lwr)) es g)
           ord kia kib sig spg sg sir spr sr
           sil spa sl (EXTP: extpub lwl spa sil),
      ksim_ esr ord kia kib (prod_gkb g sl) (sig, sir) (spg, spr) (sg, sr) <ks=
      ksim_ es (wfo_zero,ord) kia kib g (sig, (sil, sir)) (spg, (spa, spr)) (sg, (sl, sr)).
  Proof.
    ii; inversion PR; subst; econs; eauto.
    i; destruct s' as (?&?&?); rrdes EXTP0 EXTP1 EXT EXT0.
    eapply LE, KR; s; auto; [M| | |M|M]; Mdo (ss; eauto using extpub_trans).
    - eauto using prod_gkb_mon.
    - eapply vclos_unlift in RV; eapply vclos_lift; eauto using vclos_mon.
  Qed.


  Lemma gknow_left:
    forall esl es (LE: le_left_true esl es)
           (DJN: forall tn (INa: lwl.(supp) tn) (INb: lwr.(supp) tn), False)
           g (G: gknow_ (wlift (wprod lwl lwr)) es g) sr,
      gknow_ (wlift lwl) esl (prod_gka g sr).
  Proof.
    i; edestruct G; econs; i.
    - eapply prod_gka_smon; eauto; i; eapply gk_mon; eauto.
      rrdes R0; repeat (split; eauto).
    - econs; eauto.
      destruct s; ss; eapply gk_lsim; eauto.
      i; eapply lsim_left; eauto.
    
    - split; i; destruct s; ss.
      + invc H; eapply gk_name in REL; des; eauto using name_mon, prod_gka_mon.
        eapply lwr.(supp_name) in REL; exfalso; eauto using DJN.
      + econs; eauto.
        eapply gk_name; eauto.
    
  Qed.

  Lemma gknow_right:
    forall esr es (LE: le_right_true esr es)
           (DJN: forall tn (INa: lwl.(supp) tn) (INb: lwr.(supp) tn), False)
           g (G: gknow_ (wlift (wprod lwl lwr)) es g) sl,
      gknow_ (wlift lwr) esr (prod_gkb g sl).
  Proof.
    i; edestruct G; econs; i.
    - eapply prod_gkb_smon; eauto. i; eapply gk_mon; eauto.
      rrdes R0; repeat (split; eauto).
    - econs; eauto.
      destruct s; ss; eapply gk_lsim; eauto.
      i; eapply lsim_right; eauto.
    
    - split; i; destruct s; ss.
      + invc H; eapply gk_name in REL; des; eauto using name_mon, prod_gkb_mon.
        eapply lwl.(supp_name) in REL; exfalso; eauto using DJN.
      + econs; eauto.
        eapply gk_name; eauto.
    
  Qed.


  Lemma configure_update_left: forall
      (STr: stable lwr)
      g sg sl sr har hbr
      (CRr: confrel lwr (prod_gkb g sl) (sg, sr) har hbr)
      sg' (EXTg: ext _ sg' sg) sl' (EXTl: ext _ sl' sl)
      es (GK: gknow_ (wlift (wprod lwl lwr)) es g)
      (GKr: gknow (wlift lwr) (prod_gkb g sl'))
      ea' eb' hag hbg hal hbl cfa cfb ma' mb'
      (CRg: confrel gw (fun sg => prod_gka g sr (sg, sl')) sg' hag hbg)
      (CRl: confrel lwl (prod_gka g sr) (sg', sl') hal hbl)
      (RLa: real La (ea' # hag # hal # har # cfa) ma')
      (RLb: real Lb (eb' # hbg # hbl # hbr # cfb) mb')
      (CO: core ea' /\ core eb'),
    exists sr',
      <<EXTP: extpub _ sr' sr>> /\
      <<CNF: configure (conflift (wprod lwl lwr)) g (sg', (sl', sr')) true ea' eb'
                       (hag # hal # har) (hbg # hbl # hbr) cfa cfb ma' mb'>>.
  Proof.
    i; hexploit (STr _ GKr); eauto.
    { eapply confrel_mon; s; eauto using prod_gkb_mon. }
    { rewrite (cplus_swap _ hal) in RLa; rewrite (cplus_swap _ hbl) in RLb.
      rr; esplits; s; econs; eauto.
      eapply confrel_mon; eauto.
      i; invc PR; econs.
      - eapply (gk_mon GK); eauto.
        split; eauto using ext_trans; split; clean; eauto.
      - eauto.
    }
    i; des; esplits; [exact EXTP|]; econs; rewrite <-?cplus_assoc; auto.
    rr; esplits; s; rr; esplits; eauto
    ; eapply confrel_mon; eauto using prod_gka_mon.
    i; invc PR; eapply (gk_mon GK); s; eauto using ext_trans.
  Qed.

  Lemma configure_update_right: forall
      (STl: stable lwl)
      g sg sl sr hal hbl
      (CRl: confrel lwl (prod_gka g sr) (sg, sl) hal hbl)
      sg' (EXTg: ext _ sg' sg) sr' (EXTr: ext _ sr' sr)
      es (GK: gknow_ (wlift (wprod lwl lwr)) es g)
      (GKl: gknow (wlift lwl) (prod_gka g sr'))
      ea' eb' hag hbg har hbr cfa cfb ma' mb'
      (CRg: confrel gw (fun sg => prod_gkb g sl (sg, sr')) sg' hag hbg)
      (CRr: confrel lwr (prod_gkb g sl) (sg', sr') har hbr)
      (RLa: real La (ea' # hag # har # hal # cfa) ma')
      (RLb: real Lb (eb' # hbg # hbr # hbl # cfb) mb')
      (CO: core ea' /\ core eb'),
    exists sl',
      <<EXTP: extpub _ sl' sl>> /\
      <<CNF: configure (conflift (wprod lwl lwr)) g (sg', (sl', sr')) true ea' eb'
                       (hag # hal # har) (hbg # hbl # hbr) cfa cfb ma' mb'>>.
  Proof.
    i; hexploit (STl _ GKl); eauto.
    { eapply confrel_mon; s; eauto using prod_gka_mon. }
    { rewrite (cplus_swap _ har) in RLa; rewrite (cplus_swap _ hbr) in RLb.
      rr; esplits; s; econs; eauto.
      eapply confrel_mon; eauto.
      i; invc PR; econs.
      - eapply (gk_mon GK); eauto.
        split; eauto using ext_trans; split; clean; eauto.
      - eauto.
    }
    rewrite (cplus_swap _ har) in RLa; rewrite (cplus_swap _ hbr) in RLb.
    i; des; esplits; [exact EXTP|]; econs; rewrite <-?cplus_assoc; auto.
    rr; esplits; s; rr; esplits; eauto
    ; eapply confrel_mon; eauto using prod_gkb_mon.
    i; invc PR; eapply (gk_mon GK); s; eauto using ext_trans.
  Qed.


  Inductive esim_prod_lhs (esl: esimtyp (wlift lwl)) (esr: esimtyp (wlift lwr))
  : esimtyp (wlift (wprod lwl lwr)) :=
  | epl_left_true ord ka kb sig sg sil sl sir sr g ty ea eb
      (EXTP: extpub lwr sr sir)
      (IN: esl ord ka kb (prod_gka g sr) (sig, sil) (sg, sl) true ty ea eb)
    : esim_prod_lhs esl esr (ord,wfo_zero) ka kb g (sig, (sil, sir)) (sg, (sl, sr)) true ty ea eb
  | epl_right_true ord ka kb sig sg sil sl sir sr g ty ea eb
      (EXTP: extpub lwl sl sil)
      (IN: esr ord ka kb (prod_gkb g sl) (sig, sir) (sg, sr) true ty ea eb)
    : esim_prod_lhs esl esr (wfo_zero,ord) ka kb g (sig, (sil, sir)) (sg, (sl, sr)) true ty ea eb
  | epl_left_false ord ka kb sig sg sg' sil sl' sir sr g ty ea eb hra hrb
      (EXTP: extpub lwr sr sir)
      (EXTg: ext gw sg' sg)
      (CR: confrel lwr (prod_gkb g sl') (sg, sr) hra hrb)
      (IN: esl ord ka kb (prod_gka g sr) (sig, sil) (sg', sl') false ty ea eb)
    : esim_prod_lhs esl esr (ord,wfo_zero) ka kb g (sig, (sil, sir)) (sg', (sl', sr)) false ty (cplus ea hra) (cplus eb hrb)
  | epl_right_false ord ka kb sig sg sg' sil sl sir sr' g ty ea eb hla hlb
      (EXTP: extpub lwl sl sil)
      (EXTg: ext gw sg' sg)
      (CR: confrel lwl (prod_gka g sr') (sg, sl) hla hlb)
      (IN: esr ord ka kb (prod_gkb g sl) (sig, sir) (sg', sr') false ty ea eb)
    : esim_prod_lhs esl esr (wfo_zero,ord) ka kb g (sig, (sil, sir)) (sg', (sl, sr')) false ty (cplus ea hla) (cplus eb hlb).



  Lemma esim_prod:
    forall (STl: stable lwl) (STr: stable lwr)
           (DJN: forall tn (INa: lwl.(supp) tn) (INb: lwr.(supp) tn), False),
      esim_prod_lhs (esim (wlift lwl)) (esim (wlift lwr)) <e= esim (wlift (wprod lwl lwr)).
  Proof.
    pcofix CIH; i; inv PR.
    - pfold; punfold IN; rewrite clear_bote in IN; destruct IN.
      econs; ii; inv CNF; rrdes RH WHY; exploit ESIM.
      + eapply gknow_ant; [|by i; eapply pesim_incl_gres; eauto].
        eapply gknow_left; eauto; ii; eapply gres_incl_id; right; eauto using epl_left_true.
      + eapply extra_plus; [|by eauto].
        eapply lwconf_extra in WHY; des; eauto.
      + eapply extra_plus; [|by eauto].
        eapply lwconf_extra in WHY; des; eauto.
      + econs; subst; [by rr; esplits; eauto using confrel_mon|by eauto| |].
        * rewrite <-cplus_assoc; rewrite <-!cplus_assoc in MA; eauto.
        * rewrite <-cplus_assoc; rewrite <-!cplus_assoc in MB; eauto.
      + clear ESIM; intro ES; des; esplits; eauto; inversion ES; eauto.
        * { 
          econs 2; eauto; i; exploit EST; eauto.
          clear EST; i; des; (destruct s'; inv CNF0; rrdes RH; ss; des; inv ES0).
          - 
            do 6 econs; exists (s, (s0, sr)); exists false; splits; eauto.
            + left; econs; eauto.
              { instantiate (1:=(_,_)); s.
                des; eauto 10 using symprod_tc1. }
              eapply gres_incl_id; right; eapply CIH; eauto.
              eapply epl_left_false; eauto.
              * eapply confrel_mon; eauto using prod_gkb_mon.
              * eapply pesim_incl_gres in ES1; eauto.
            + econs; eauto.
              * rewrite <-cplus_assoc, (cplus_swap _ hay0); eauto.
              * rewrite <-cplus_assoc, (cplus_swap _ hby0); eauto.
          - 
            rrdes RH; subst; rewrite <-cplus_assoc in MA0, MB0.
            hexploit configure_update_left; eauto.
            { eapply gknow_right; eauto; ii; eapply gres_incl_id; right.
              eapply CIH; eauto using epl_right_true.
            }
            i; des.
            do 6 econs; exists (s, (s0, sr')), true.
            inv SU; ss; splits; des; eauto.
            + 
              right; econs; eauto; econs 1; [eapply APPa|eapply APPb|inversion RF| |]; eauto.
              * invc VR; econs 1; eauto; eapply (gk_mon GK); s; eauto using ext_trans.
              
              * eapply vclos_lift, vclos_unlift; eapply vclos_mon; [eapply RV| |]; s; eauto.
                i; invc PR0; eapply (gk_mon GK); s; eauto using ext_trans. 
              * eapply ksim_mon
                ; first (by instantiate (4:=(_,_))
                         ; eapply ksim_left; eauto; ii; eapply gres_incl_id
                         ; right; eapply pesim_incl_gres in H; eapply CIH
                         ; eauto using epl_left_true)
                ; s; eauto.
            + 
              right; econs; eauto; econs 2; [eapply INSTa|eapply INSTb|inversion RF| |]; eauto.
              
              * invc VR; econs 2; eauto; eapply (gk_mon GK); s; eauto using ext_trans.
              * eapply ksim_mon
                ; first (by instantiate (4:=(_,_))
                         ; eapply ksim_left; eauto; ii; eapply gres_incl_id
                         ; right; eapply pesim_incl_gres in H; eapply CIH
                         ; eauto using epl_left_true)
                ; s; eauto.
          }
        * { 
            inv CNF0; rrdes RH; destruct s'; des.
            move CNF0 at bottom; move MA0 at bottom; move MB0 at bottom; subst.
            rewrite <-!cplus_assoc in MA0, MB0.
            dup EXT; rrdes EXT; dup EXTP0; rrdes EXTP0.
            hexploit configure_update_left; eauto.
            { eapply gknow_right; eauto; ii; eapply gres_incl_id; right.
              eapply CIH; eauto using epl_right_true.
            }
            i; des.
            eapply esim_ret with (s':=(s,(s0,sr'))); first 2 last; last first; try by (ss; eauto).
            - ss; des; splits; eauto using extpub_trans.
            - eapply vclos_mon; [by eapply vclos_lift, vclos_unlift; eauto|by eauto|].
              i; invc PR0; eapply (gk_mon GK); s; eauto using ext_trans.
          }
    - 
      pfold; punfold IN; rewrite clear_bote in IN; destruct IN.
      econs; ii; inv CNF; rrdes RH WHY; exploit ESIM.
      + eapply gknow_ant; [|by i; eapply pesim_incl_gres; eauto].
        eapply gknow_right; eauto; ii; eapply gres_incl_id; right; eauto using epl_right_true.
      + eapply extra_plus; [|by eauto].
        eapply lwconf_extra in WHX0; des; eauto.
      + eapply extra_plus; [|by eauto].
        eapply lwconf_extra in WHX0; des; eauto.
      + econs; subst; [by rr; esplits; eauto using confrel_mon|by eauto| |].
        * rewrite <-!cplus_assoc in MA.
          rewrite <-cplus_assoc, (cplus_swap _ hay0); eauto.
        * rewrite <-!cplus_assoc in MB.
          rewrite <-cplus_assoc, (cplus_swap _ hby0); eauto.
      + clear ESIM; intro ES; des; esplits; eauto; inversion ES; eauto.
        * { 
          econs 2; eauto; i; exploit EST; eauto.
          clear EST; i; des; (destruct s'; inv CNF0; rrdes RH; ss; des; inv ES0).
          - 
            do 6 econs; exists (s, (sl, s0)); exists false; splits; eauto.
            + left; econs; eauto.
              { instantiate (1:=(_,_)); s.
                des; eauto 10 using symprod_tc2. }
              eapply gres_incl_id; right; eapply CIH; eauto.
              eapply epl_right_false; eauto.
              * eapply confrel_mon; eauto using prod_gka_mon.
              * eapply pesim_incl_gres in ES1; eauto.
            + econs; eauto.
              * rewrite <-cplus_assoc, (cplus_swap _ hax0); eauto.
              * rewrite <-cplus_assoc, (cplus_swap _ hbx0); eauto.
          - 
            rrdes RH; subst; rewrite <-cplus_assoc in MA0, MB0.
            hexploit configure_update_right; eauto.
            { eapply gknow_left; eauto; ii; eapply gres_incl_id; right.
              eapply CIH; eauto using epl_left_true.
            }
            i; des.
            do 6 econs; exists (s, (sl', s0)), true.
            inv SU; ss; splits; des; eauto.
            + 
              right; econs; eauto; econs 1; [eapply APPa|eapply APPb|inversion RF| |]; eauto.
              * invc VR; econs 1; eauto; eapply (gk_mon GK); s; eauto using ext_trans.
              
              * eapply vclos_lift, vclos_unlift; eapply vclos_mon; [eapply RV| |]; s; eauto.
                i; invc PR0; eapply (gk_mon GK); s; eauto using ext_trans.
              * eapply ksim_mon
                ; first (by instantiate (4:=(_,_))
                         ; eapply ksim_right; eauto; ii; eapply gres_incl_id
                         ; right; eapply pesim_incl_gres in H; eapply CIH
                         ; eauto using epl_right_true)
                ; s; eauto.
            + 
              right; econs; eauto; econs 2; [eapply INSTa|eapply INSTb|inversion RF| |]; eauto.
              
              * invc VR; econs 2; eauto; eapply (gk_mon GK); s; eauto using ext_trans.
              * eapply ksim_mon
                ; first (by instantiate (4:=(_,_))
                         ; eapply ksim_right; eauto; ii; eapply gres_incl_id
                         ; right; eapply pesim_incl_gres in H; eapply CIH
                         ; eauto using epl_right_true)
                ; s; eauto.
          }
        * { 
            inv CNF0; rrdes RH; destruct s'; des.
            move CNF0 at bottom; move MA0 at bottom; move MB0 at bottom; subst.
            rewrite <-!cplus_assoc in MA0, MB0.
            dup EXT; rrdes EXT; dup EXTP0; rrdes EXTP0.
            hexploit configure_update_right; eauto.
            { eapply gknow_left; eauto; ii; eapply gres_incl_id; right.
              eapply CIH; eauto using epl_left_true.
            }
            i; des.
            eapply esim_ret with (s':=(s,(sl', s0))); first 2 last; last first; try by (ss; eauto).
            - ss; des; splits; eauto using extpub_trans.
            - eapply vclos_mon; [by eapply vclos_lift, vclos_unlift; eauto|by eauto|].
              i; invc PR0; eapply (gk_mon GK); s; eauto using ext_trans.
          }
    - pfold; punfold IN; rewrite clear_bote in IN; destruct IN.
      econs; ii; inv CNF; des; subst; noemp; exploit ESIM.
      + eapply gknow_ant; [|by i; eapply pesim_incl_gres; eauto].
        eapply gknow_left; eauto; ii; eapply gres_incl_id; right; eauto using epl_left_true.
      + eapply extra_plus; [|by eauto].
        eapply lwconf_extra in CR; des; clear XTRa; eauto.
      + eapply extra_plus; [|by eauto].
        eapply lwconf_extra in CR; des; clear XTRb; eauto.
      + rewrite <-cplus_assoc in MA; rewrite <-cplus_assoc in MB; econs; eauto; noemp; eauto.
      + clear ESIM; intro ES; des; esplits; eauto; inversion ES; eauto.
        * { 
          econs 2; eauto; i; exploit EST; eauto.
          clear EST; i; des; (destruct s'; inv CNF0; rrdes RH; ss; des; inv ES0).
          - 
            do 6 econs; exists (s, (s0, sr)); exists false; splits; eauto.
            + left; econs; eauto.
              { instantiate (1:=(_,_)); s.
                des; eauto 10 using symprod_tc1. }
              eapply gres_incl_id; right; eapply CIH; eauto.
              eapply epl_left_false
              ; [| |by eapply confrel_mon; eauto using prod_gkb_mon|]
              ; eauto using ext_trans.
              eapply pesim_incl_gres in ES1; eauto.
            + econs; eauto.
              * rewrite <-cplus_assoc, (cplus_swap _ hra); eauto.
              * rewrite <-cplus_assoc, (cplus_swap _ hrb); eauto.
          - 
            rrdes RH; subst; rewrite <-cplus_assoc in MA0, MB0.
            hexploit configure_update_left; [| |eapply ext_trans| | | | | | | | |]; eauto.
            { eapply gknow_right; eauto; ii; eapply gres_incl_id; right.
              eapply CIH; eauto using epl_right_true.
            }
            i; des.
            do 6 econs; exists (s, (s0, sr')), true.
            inv SU; ss; splits; des; eauto.
            + 
              right; econs; eauto; econs 1; [eapply APPa|eapply APPb|inversion RF| |]; eauto.
              * invc VR; econs 1; eauto; eapply (gk_mon GK); s; eauto using ext_trans.
              
              * eapply vclos_lift, vclos_unlift; eapply vclos_mon; [eapply RV| |]; s; eauto.
                i; invc PR0; eapply (gk_mon GK); s; eauto using ext_trans.
              * eapply ksim_mon
                ; first (by instantiate (4:=(_,_)) 
                         ; eapply ksim_left; eauto; ii; eapply gres_incl_id
                         ; right; eapply pesim_incl_gres in H; eapply CIH
                         ; eauto using epl_left_true)
                ; s; eauto.
            + 
              right; econs; eauto; econs 2; [eapply INSTa|eapply INSTb|inversion RF| |]; eauto.
              
              * invc VR; econs 2; eauto; eapply (gk_mon GK); s; eauto using ext_trans.
              * eapply ksim_mon
                ; first (by instantiate (4:=(_,_)) 
                         ; eapply ksim_left; eauto; ii; eapply gres_incl_id
                         ; right; eapply pesim_incl_gres in H; eapply CIH
                         ; eauto using epl_left_true)
                ; s; eauto.
          }
        * { 
            inv CNF0; rrdes RH; destruct s'; des.
            move CNF0 at bottom; move MA0 at bottom; move MB0 at bottom; subst.
            rewrite <-!cplus_assoc in MA0, MB0.
            dup EXT; rrdes EXT; dup EXTP0; rrdes EXTP0.
            hexploit configure_update_left; [| |eapply ext_trans| | | | | | | | |]; eauto.
            { eapply gknow_right; eauto; ii; eapply gres_incl_id; right.
              eapply CIH; eauto using epl_right_true.
            }
            i; des.
            eapply esim_ret with (s':=(s,(s0,sr'))); first 2 last; last first; try by (ss; eauto).
            - ss; des; splits; eauto using extpub_trans.
            - eapply vclos_mon; [by eapply vclos_lift, vclos_unlift; eauto|by eauto|].
              i; invc PR0; eapply (gk_mon GK); s; eauto using ext_trans.
          }
    - 
      pfold; punfold IN; rewrite clear_bote in IN; destruct IN.
      econs; ii; inv CNF; des; subst; noemp; exploit ESIM.
      + eapply gknow_ant; [|by i; eapply pesim_incl_gres; eauto].
        eapply gknow_right; eauto; ii; eapply gres_incl_id; right; eauto using epl_right_true.
      + eapply extra_plus; [|by eauto].
        eapply lwconf_extra in CR; des; clear XTRa; eauto.
      + eapply extra_plus; [|by eauto].
        eapply lwconf_extra in CR; des; clear XTRb; eauto.
      + rewrite <-cplus_assoc in MA; rewrite <-cplus_assoc in MB; econs; eauto; noemp; eauto.
      + clear ESIM; intro ES; des; esplits; eauto; inversion ES; eauto.
        * { 
          econs 2; eauto; i; exploit EST; eauto.
          clear EST; i; des; (destruct s'; inv CNF0; rrdes RH; ss; des; inv ES0).
          - 
            do 6 econs; exists (s, (sl, s0)); exists false; splits; eauto.
            + left; econs; eauto.
              { instantiate (1:=(_,_)); s.
                des; eauto 10 using symprod_tc2. }
              eapply gres_incl_id; right; eapply CIH; eauto.
              eapply epl_right_false
              ; [| |by eapply confrel_mon; eauto using prod_gka_mon|]
              ; eauto using ext_trans.
              eapply pesim_incl_gres in ES1; eauto.
            + econs; eauto.
              * rewrite <-cplus_assoc, (cplus_swap _ hla); eauto.
              * rewrite <-cplus_assoc, (cplus_swap _ hlb); eauto.
          - 
            rrdes RH; subst; rewrite <-cplus_assoc in MA0, MB0.
            hexploit configure_update_right; [| |eapply ext_trans| | | | | | | | |]; eauto.
            { eapply gknow_left; eauto; ii; eapply gres_incl_id; right.
              eapply CIH; eauto using epl_left_true.
            }
            i; des.
            do 6 econs; exists (s, (sl', s0)), true.
            inv SU; ss; splits; des; eauto.
            + 
              right; econs; eauto; econs 1; [eapply APPa|eapply APPb|inversion RF| |]; eauto.
              * invc VR; econs 1; eauto; eapply (gk_mon GK); s; eauto using ext_trans.
              
              * eapply vclos_lift, vclos_unlift; eapply vclos_mon; [eapply RV| |]; s; eauto.
                i; invc PR0; eapply (gk_mon GK); s; eauto using ext_trans.
              * eapply ksim_mon
                ; first (by instantiate (4:=(_,_))
                         ; eapply ksim_right; eauto; ii; eapply gres_incl_id
                         ; right; eapply pesim_incl_gres in H; eapply CIH
                         ; eauto using epl_right_true)
                ; s; eauto.
            + 
              right; econs; eauto; econs 2; [eapply INSTa|eapply INSTb|inversion RF| |]; eauto.
              
              * invc VR; econs 2; eauto; eapply (gk_mon GK); s; eauto using ext_trans.
              * eapply ksim_mon
                ; first (by instantiate (4:=(_,_))
                         ; eapply ksim_right; eauto; ii; eapply gres_incl_id
                         ; right; eapply pesim_incl_gres in H; eapply CIH
                         ; eauto using epl_right_true)
                ; s; eauto.
          }
        * { 
            inv CNF0; rrdes RH; destruct s'; des.
            move CNF0 at bottom; move MA0 at bottom; move MB0 at bottom; subst.
            rewrite <-!cplus_assoc in MA0, MB0.
            dup EXT; rrdes EXT; dup EXTP0; rrdes EXTP0.
            hexploit configure_update_right; [| |eapply ext_trans| | | | | | | | |]; eauto.
            { eapply gknow_left; eauto; ii; eapply gres_incl_id; right.
              eapply CIH; eauto using epl_left_true.
            }
            i; des.
            eapply esim_ret with (s':=(s,(sl',s0))); first 2 last; last first; try by (ss; eauto).
            - ss; des; splits; eauto using extpub_trans.
            - eapply vclos_mon; [by eapply vclos_lift, vclos_unlift; eauto|by eauto|].
              i; invc PR0; eapply (gk_mon GK); s; eauto using ext_trans.
          }
  Qed.


  Lemma gknow_prod:
    forall (STa: stable lwl) (STb: stable lwr)
           (DJN: forall tn (INa: lwl.(supp) tn) (INb: lwr.(supp) tn), False)
           g (G: gknow (wlift (wprod lwl lwr)) g),
      (forall sr, gknow (wlift lwl) (prod_gka g sr)) /\
      (forall sl, gknow (wlift lwr) (prod_gkb g sl)).
  Proof.
    i; split; i; [eapply gknow_left|eapply gknow_right]; eauto
    ; ii; eapply esim_prod; eauto using epl_left_true, epl_right_true.
  Qed.


  Lemma stable_prod:
    forall (STa: stable lwl) (STb: stable lwr)
           (DJN: forall tn (INa: lwl.(supp) tn) (INb: lwr.(supp) tn), False),
      stable (wprod lwl lwr).
  Proof.
    ii; rrdes CR; destruct s; subst; ss.

    set (ga s' := g (fst s', (snd s', s0))).
    set (gb s' := g (fst s', (s, snd s'))).
    change (g _) with (ga (sp, s)) in WHX.
    change (g _) with (gb (sp, s0)) in WHY.

    eapply STa in WHX; [|by apply gknow_prod; eauto]; exploit WHX; eauto.
    { rrdes COMPAT; inversion COMPAT; rr; esplits; econs; eauto using confrel_mon.
      - rewrite <-cplus_assoc in MA; eauto.
      - rewrite <-cplus_assoc in MB; eauto.
    }
    i; des.

    eapply STb in WHY; [|by apply gknow_prod; eauto]; exploit WHY; eauto.
    { rrdes COMPAT; inversion COMPAT; rr; esplits; econs; eauto using confrel_mon.
      - rewrite <-cplus_assoc, !(cplus_swap _ hax) in MA; eauto.
      - rewrite <-cplus_assoc, !(cplus_swap _ hbx) in MB; eauto.
    }
    i; des.

    exists (s', s'0).
    unfold ga, gb in *; rr; esplits; eauto; s.
    rr; esplits; eauto; eapply confrel_mon
    ; eauto using prod_gka_mon, prod_gkb_mon.
  Qed.

End esim.
