(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program Relations Omega.
Require Import paco2  paco10  vgtac.
Require Import basis lang_common lang_spec.


Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



Fixpoint dmach L n : Type :=
  match n with
    | O => L.(mach)
    | S n' => L.(conf) -> dmach L n'
  end.

Inductive dance {L} (m: L.(mach)): forall n, dmach L n -> Prop :=
  | dance_O
    : forall n (EQ: m = n), dance m O n
  | dance_S (dm': L.(conf) -> L.(mach)) n (dm'': dmach L (S n))
            (ST: forall cf (XTR: extra cf) mm (MM: L.(mcplus) m cf mm),
                 exists mm', <<ST: L.(step) etau mm mm'>> /\
                             <<DET: forall evt mm'' (ST': L.(step) evt mm mm''),
                                      evt = etau /\ mm'' = mm'>> /\
                             <<MM': L.(mcplus) (dm' cf) cf mm'>> /\
                             <<DA: dance (dm' cf) _ (dm'' cf)>>)
    : dance m (S n) dm''
.

Inductive vector A : nat -> Type :=
  | vnil: vector A 0
  | vcons: forall (hd: A) n (tl: vector A n), vector A (S n).

Definition vhd A n (v: vector A (S n)) : A :=
  match v with
    | vcons a _ _ => a
  end.

Definition vtl A n (v: vector A (S n)) : vector A n :=
  match v with
    | vcons _ _ v' => v'
  end.

Inductive vIn {A} (a:A): forall {n}, vector A n -> Prop :=
 | In_cons_hd {m} (v: vector A m): vIn a (vcons a v)
 | In_cons_tl {m} x (v: vector A m): vIn a v -> vIn a (vcons x v).
Hint Constructors vIn.


Fixpoint dmach_apply {L} n (dm: dmach L n) (v: vector L.(conf) n) : L.(mach) :=
  match n return dmach L n -> vector L.(conf) n -> L.(mach) with
    | O => fun dm _ => dm
    | S n' => fun dm v => dmach_apply (dm (vhd v)) (vtl v)
  end dm v.
