(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

open Parser
open Pils_tgt
open Pilsner


let compile sel files : unit =
  let cin = open_in (List.nth files 0) in
  let lexbuf = Lexing.from_channel cin in
  let (imports, msrc) = Parser.annotated_modul Lexer.token lexbuf in
  let mtgt = pilsner sel imports msrc in
  save_module mtgt (List.nth files 1)
;;


let _ =
  let sel = ref do_all in
  let flags =
    [("-no-inline", Arg.Unit (fun _ -> sel := {!sel with do_inline = false}), " Disable inlining.")
    ;("-no-contify", Arg.Unit (fun _ -> sel := {!sel with do_contify = false}), " Disable contification.")
    ;("-no-dce", Arg.Unit (fun _ -> sel := {!sel with do_dce = false}), " Disable dead code elimination.")
    ;("-no-hoist", Arg.Unit (fun _ -> sel := {!sel with do_hoist = false}), " Disable hoisting.")
    ;("-no-commute", Arg.Unit (fun _ -> sel := {!sel with do_commute = false}), " Disable commutation.")
    ;("-no-dedup", Arg.Unit (fun _ -> sel := {!sel with do_dedup = false}), " Disable deduplication.")
    ] in
  let files = ref [] in
  Arg.parse flags (fun s -> files := !files @ [s]) "";
  compile !sel !files
;;
